# Websocket Request/Response Handler
#
# Copyright (C) 2020 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import logging
import asyncio
from tornado.websocket import WebSocketHandler, WebSocketClosedError
from tornado.web import HTTPError
from ..common import (
    RequestType,
    WebRequest,
    BaseRemoteConnection,
    TransportType,
)
from ..utils import ServerError, parse_ip_address

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Callable,
    Coroutine,
    Tuple,
    Union,
    Dict,
    List,
)

if TYPE_CHECKING:
    from ..server import Server
    from .klippy_connection import KlippyConnection as Klippy
    from ..confighelper import ConfigHelper
    from .application import MoonrakerApp
    from .extensions import ExtensionManager
    from .authorization import Authorization
    from ..utils import IPAddress
    ConvType = Union[str, bool, float, int]
    ArgVal = Union[None, int, float, bool, str]
    RPCCallback = Callable[..., Coroutine]
    AuthComp = Optional[Authorization]

CLIENT_TYPES = ["web", "mobile", "desktop", "display", "bot", "agent", "other"]

class WebsocketManager:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.clients: Dict[int, BaseRemoteConnection] = {}
        self.bridge_connections: Dict[int, BridgeSocket] = {}
        self.closed_event: Optional[asyncio.Event] = None
        app: MoonrakerApp = self.server.lookup_component("application")
        app.register_websocket_handler("/websocket", WebSocket)
        app.register_websocket_handler("/klippysocket", BridgeSocket)
        self.server.register_endpoint(
            "/server/websocket/id", RequestType.GET, self._handle_id_request,
            TransportType.WEBSOCKET
        )
        self.server.register_endpoint(
            "/server/connection/identify", RequestType.POST, self._handle_identify,
            TransportType.WEBSOCKET, auth_required=False
        )

    def register_notification(
        self,
        event_name: str,
        notify_name: Optional[str] = None,
        event_type: Optional[str] = None
    ) -> None:
        if notify_name is None:
            notify_name = event_name.split(':')[-1]
        if event_type == "logout":
            def notify_handler(*args):
                self.notify_clients(notify_name, args)
                self._process_logout(*args)
        else:
            def notify_handler(*args):
                self.notify_clients(notify_name, args)
        self.server.register_event_handler(event_name, notify_handler)

    async def _handle_id_request(self, web_request: WebRequest) -> Dict[str, int]:
        sc = web_request.get_client_connection()
        assert sc is not None
        return {'websocket_id': sc.uid}

    async def _handle_identify(self, web_request: WebRequest) -> Dict[str, int]:
        sc = web_request.get_client_connection()
        assert sc is not None
        if sc.identified:
            raise self.server.error(
                f"Connection already identified: {sc.client_data}"
            )
        name = web_request.get_str("client_name")
        version = web_request.get_str("version")
        client_type: str = web_request.get_str("type").lower()
        url = web_request.get_str("url")
        sc.authenticate(
            token=web_request.get_str("access_token", None),
            api_key=web_request.get_str("api_key", None)
        )

        if client_type not in CLIENT_TYPES:
            raise self.server.error(f"Invalid Client Type: {client_type}")
        sc.client_data = {
            "name": name,
            "version": version,
            "type": client_type,
            "url": url
        }
        if client_type == "agent":
            extensions: ExtensionManager
            extensions = self.server.lookup_component("extensions")
            try:
                extensions.register_agent(sc)
            except ServerError:
                sc.client_data["type"] = ""
                raise
        logging.info(
            f"Websocket {sc.uid} Client Identified - "
            f"Name: {name}, Version: {version}, Type: {client_type}"
        )
        self.server.send_event("websockets:client_identified", sc)
        return {'connection_id': sc.uid}

    def _process_logout(self, user: Dict[str, Any]) -> None:
        if "username" not in user:
            return
        name = user["username"]
        for sc in self.clients.values():
            sc.on_user_logout(name)

    def has_socket(self, ws_id: int) -> bool:
        return ws_id in self.clients

    def get_client(self, uid: int) -> Optional[BaseRemoteConnection]:
        return self.clients.get(uid, None)

    def get_client_ws(self, ws_id: int) -> Optional[WebSocket]:
        sc = self.clients.get(ws_id, None)
        if sc is None or not isinstance(sc, WebSocket):
            return None
        return sc

    def get_clients_by_type(
        self, client_type: str
    ) -> List[BaseRemoteConnection]:
        if not client_type:
            return []
        ret: List[BaseRemoteConnection] = []
        for sc in self.clients.values():
            if sc.client_data.get("type", "") == client_type.lower():
                ret.append(sc)
        return ret

    def get_clients_by_name(self, name: str) -> List[BaseRemoteConnection]:
        if not name:
            return []
        ret: List[BaseRemoteConnection] = []
        for sc in self.clients.values():
            if sc.client_data.get("name", "").lower() == name.lower():
                ret.append(sc)
        return ret

    def get_unidentified_clients(self) -> List[BaseRemoteConnection]:
        ret: List[BaseRemoteConnection] = []
        for sc in self.clients.values():
            if not sc.client_data:
                ret.append(sc)
        return ret

    def add_client(self, sc: BaseRemoteConnection) -> None:
        self.clients[sc.uid] = sc
        self.server.send_event("websockets:client_added", sc)
        logging.debug(f"New Websocket Added: {sc.uid}")

    def remove_client(self, sc: BaseRemoteConnection) -> None:
        old_sc = self.clients.pop(sc.uid, None)
        if old_sc is not None:
            self.server.send_event("websockets:client_removed", sc)
            logging.debug(f"Websocket Removed: {sc.uid}")
        self._check_closed_event()

    def add_bridge_connection(self, bc: BridgeSocket) -> None:
        self.bridge_connections[bc.uid] = bc
        logging.debug(f"New Bridge Connection Added: {bc.uid}")

    def remove_bridge_connection(self, bc: BridgeSocket) -> None:
        old_bc = self.bridge_connections.pop(bc.uid, None)
        if old_bc is not None:
            logging.debug(f"Bridge Connection Removed: {bc.uid}")
        self._check_closed_event()

    def _check_closed_event(self) -> None:
        if (
            self.closed_event is not None and
            not self.clients and
            not self.bridge_connections
        ):
            self.closed_event.set()

    def notify_clients(
        self,
        name: str,
        data: Union[List, Tuple] = [],
        mask: List[int] = []
    ) -> None:
        msg: Dict[str, Any] = {'jsonrpc': "2.0", 'method': "notify_" + name}
        if data:
            msg['params'] = data
        for sc in list(self.clients.values()):
            if sc.uid in mask or sc.need_auth:
                continue
            sc.queue_message(msg)

    def get_count(self) -> int:
        return len(self.clients)

    async def close(self) -> None:
        if not self.clients:
            return
        self.closed_event = asyncio.Event()
        for bc in list(self.bridge_connections.values()):
            bc.close_socket(1001, "Server Shutdown")
        for sc in list(self.clients.values()):
            sc.close_socket(1001, "Server Shutdown")
        try:
            await asyncio.wait_for(self.closed_event.wait(), 2.)
        except asyncio.TimeoutError:
            pass
        self.closed_event = None

class WebSocket(WebSocketHandler, BaseRemoteConnection):
    connection_count: int = 0

    def initialize(self) -> None:
        self.on_create(self.settings['server'])
        self._ip_addr = parse_ip_address(self.request.remote_ip or "")
        self.last_pong_time: float = self.eventloop.get_loop_time()
        self.cors_allowed: bool = False

    @property
    def ip_addr(self) -> Optional[IPAddress]:
        return self._ip_addr

    @property
    def hostname(self) -> str:
        return self.request.host_name

    def get_current_user(self) -> Any:
        return self._user_info

    def open(self, *args, **kwargs) -> None:
        self.__class__.connection_count += 1
        self.set_nodelay(True)
        self._connected_time = self.eventloop.get_loop_time()
        agent = self.request.headers.get("User-Agent", "")
        is_proxy = False
        if (
            "X-Forwarded-For" in self.request.headers or
            "X-Real-Ip" in self.request.headers
        ):
            is_proxy = True
        logging.info(f"Websocket Opened: ID: {self.uid}, "
                     f"Proxied: {is_proxy}, "
                     f"User Agent: {agent}, "
                     f"Host Name: {self.hostname}")
        self.wsm.add_client(self)

    def on_message(self, message: Union[bytes, str]) -> None:
        self.eventloop.register_callback(self._process_message, message)

    def on_pong(self, data: bytes) -> None:
        self.last_pong_time = self.eventloop.get_loop_time()

    def on_close(self) -> None:
        self.is_closed = True
        self.__class__.connection_count -= 1
        kconn: Klippy = self.server.lookup_component("klippy_connection")
        kconn.remove_subscription(self)
        self.message_buf = []
        now = self.eventloop.get_loop_time()
        pong_elapsed = now - self.last_pong_time
        for resp in self.pending_responses.values():
            resp.set_exception(ServerError("Client Socket Disconnected", 500))
        self.pending_responses = {}
        logging.info(f"Websocket Closed: ID: {self.uid} "
                     f"Close Code: {self.close_code}, "
                     f"Close Reason: {self.close_reason}, "
                     f"Pong Time Elapsed: {pong_elapsed:.2f}")
        if self._client_data["type"] == "agent":
            extensions: ExtensionManager
            extensions = self.server.lookup_component("extensions")
            extensions.remove_agent(self)
        self.wsm.remove_client(self)

    async def write_to_socket(self, message: Union[bytes, str]) -> None:
        try:
            await self.write_message(message)
        except WebSocketClosedError:
            self.is_closed = True
            self.message_buf.clear()
            logging.info(
                f"Websocket closed while writing: {self.uid}")
        except Exception:
            logging.exception(
                f"Error sending data over websocket: {self.uid}")

    def check_origin(self, origin: str) -> bool:
        if not super(WebSocket, self).check_origin(origin):
            return self.cors_allowed
        return True

    def on_user_logout(self, user: str) -> bool:
        if super().on_user_logout(user):
            self._need_auth = True
            return True
        return False

    # Check Authorized User
    async def prepare(self) -> None:
        max_conns = self.settings["max_websocket_connections"]
        if self.__class__.connection_count >= max_conns:
            raise self.server.error(
                "Maximum Number of Websocket Connections Reached"
            )
        auth: AuthComp = self.server.lookup_component('authorization', None)
        if auth is not None:
            try:
                self._user_info = await auth.authenticate_request(self.request)
            except Exception as e:
                logging.info(f"Websocket Failed Authentication: {e}")
                self._user_info = None
                self._need_auth = True
            if "Origin" in self.request.headers:
                origin = self.request.headers.get("Origin")
            else:
                origin = self.request.headers.get("Sec-Websocket-Origin", None)
            self.cors_allowed = await auth.check_cors(origin)

    def close_socket(self, code: int, reason: str) -> None:
        self.close(code, reason)

class BridgeSocket(WebSocketHandler):
    def initialize(self) -> None:
        self.server: Server = self.settings['server']
        self.wsm: WebsocketManager = self.server.lookup_component("websockets")
        self.eventloop = self.server.get_event_loop()
        self.uid = id(self)
        self._ip_addr = parse_ip_address(self.request.remote_ip or "")
        self.last_pong_time: float = self.eventloop.get_loop_time()
        self.is_closed = False
        self.klippy_writer: Optional[asyncio.StreamWriter] = None
        self.klippy_write_buf: List[bytes] = []
        self.klippy_queue_busy: bool = False
        self.cors_allowed: bool = False

    @property
    def ip_addr(self) -> Optional[IPAddress]:
        return self._ip_addr

    @property
    def hostname(self) -> str:
        return self.request.host_name

    def open(self, *args, **kwargs) -> None:
        WebSocket.connection_count += 1
        self.set_nodelay(True)
        self._connected_time = self.eventloop.get_loop_time()
        agent = self.request.headers.get("User-Agent", "")
        is_proxy = False
        if (
            "X-Forwarded-For" in self.request.headers or
            "X-Real-Ip" in self.request.headers
        ):
            is_proxy = True
        logging.info(f"Bridge Socket Opened: ID: {self.uid}, "
                     f"Proxied: {is_proxy}, "
                     f"User Agent: {agent}, "
                     f"Host Name: {self.hostname}")
        self.wsm.add_bridge_connection(self)

    def on_message(self, message: Union[bytes, str]) -> None:
        if isinstance(message, str):
            message = message.encode(encoding="utf-8")
        self.klippy_write_buf.append(message)
        if self.klippy_queue_busy:
            return
        self.klippy_queue_busy = True
        self.eventloop.register_callback(self._write_klippy_messages)

    async def _write_klippy_messages(self) -> None:
        while self.klippy_write_buf:
            if self.klippy_writer is None or self.is_closed:
                break
            msg = self.klippy_write_buf.pop(0)
            try:
                self.klippy_writer.write(msg + b"\x03")
                await self.klippy_writer.drain()
            except asyncio.CancelledError:
                raise
            except Exception:
                if not self.is_closed:
                    logging.debug("Klippy Disconnection From _write_request()")
                    self.close(1001, "Klippy Disconnected")
                break
        self.klippy_queue_busy = False

    def on_pong(self, data: bytes) -> None:
        self.last_pong_time = self.eventloop.get_loop_time()

    def on_close(self) -> None:
        WebSocket.connection_count -= 1
        self.is_closed = True
        self.klippy_write_buf.clear()
        if self.klippy_writer is not None:
            self.klippy_writer.close()
            self.klippy_writer = None
        now = self.eventloop.get_loop_time()
        pong_elapsed = now - self.last_pong_time
        logging.info(f"Bridge Socket Closed: ID: {self.uid} "
                     f"Close Code: {self.close_code}, "
                     f"Close Reason: {self.close_reason}, "
                     f"Pong Time Elapsed: {pong_elapsed:.2f}")
        self.wsm.remove_bridge_connection(self)

    async def _read_unix_stream(self, reader: asyncio.StreamReader) -> None:
        errors_remaining: int = 10
        while not reader.at_eof():
            try:
                data = memoryview(await reader.readuntil(b'\x03'))
            except (ConnectionError, asyncio.IncompleteReadError):
                break
            except asyncio.CancelledError:
                logging.exception("Klippy Stream Read Cancelled")
                raise
            except Exception:
                logging.exception("Klippy Stream Read Error")
                errors_remaining -= 1
                if not errors_remaining or self.is_closed:
                    break
                continue
            try:
                await self.write_message(data[:-1].tobytes())
            except WebSocketClosedError:
                logging.info(
                    f"Bridge closed while writing: {self.uid}")
                break
            except asyncio.CancelledError:
                raise
            except Exception:
                logging.exception(
                    f"Error sending data over Bridge: {self.uid}")
                errors_remaining -= 1
                if not errors_remaining or self.is_closed:
                    break
                continue
            errors_remaining = 10
        if not self.is_closed:
            logging.debug("Bridge Disconnection From _read_unix_stream()")
            self.close_socket(1001, "Klippy Disconnected")

    def check_origin(self, origin: str) -> bool:
        if not super().check_origin(origin):
            return self.cors_allowed
        return True

    # Check Authorized User
    async def prepare(self) -> None:
        max_conns = self.settings["max_websocket_connections"]
        if WebSocket.connection_count >= max_conns:
            raise self.server.error(
                "Maximum Number of Bridge Connections Reached"
            )
        auth: AuthComp = self.server.lookup_component("authorization", None)
        if auth is not None:
            self.current_user = await auth.authenticate_request(self.request)
            if "Origin" in self.request.headers:
                origin = self.request.headers.get("Origin")
            else:
                origin = self.request.headers.get("Sec-Websocket-Origin", None)
            self.cors_allowed = await auth.check_cors(origin)
        kconn: Klippy = self.server.lookup_component("klippy_connection")
        try:
            reader, writer = await kconn.open_klippy_connection()
        except ServerError as err:
            raise HTTPError(err.status_code, str(err)) from None
        except Exception as e:
            raise HTTPError(503, "Failed to open connection to Klippy") from e
        self.klippy_writer = writer
        self.eventloop.register_callback(self._read_unix_stream, reader)

    def close_socket(self, code: int, reason: str) -> None:
        self.close(code, reason)

def load_component(config: ConfigHelper) -> WebsocketManager:
    return WebsocketManager(config)



# MQTT client implementation for Moonraker
#
# Copyright (C) 2021  Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import socket
import asyncio
import logging
import pathlib
import ssl
from collections import deque
import paho.mqtt.client as paho_mqtt
import paho.mqtt
from ..common import (
    TransportType,
    RequestType,
    WebRequest,
    APITransport,
    KlippyState
)
from ..utils import json_wrapper as jsonw

# Annotation imports
from typing import (
    List,
    Optional,
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Dict,
    Union,
    Tuple,
    Awaitable,
    Deque,
)
if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import JsonRPC, APIDefinition
    from ..eventloop import FlexTimer
    from .klippy_apis import KlippyAPI
    FlexCallback = Callable[[bytes], Optional[Coroutine]]
    RPCCallback = Callable[..., Coroutine]

PAHO_MQTT_VERSION = tuple([int(p) for p in paho.mqtt.__version__.split(".")])
DUP_API_REQ_CODE = -10000
MQTT_PROTOCOLS = {
    'v3.1': paho_mqtt.MQTTv31,
    'v3.1.1': paho_mqtt.MQTTv311,
    'v5': paho_mqtt.MQTTv5
}

class ExtPahoClient(paho_mqtt.Client):
    # Override reconnection to take a connected socket.  This allows Moonraker
    # create the socket connection asynchronously
    def reconnect(self, sock: Optional[socket.socket] = None):
        """Reconnect the client after a disconnect. Can only be called after
        connect()/connect_async()."""
        if len(self._host) == 0:
            raise ValueError('Invalid host.')
        if self._port <= 0:
            raise ValueError('Invalid port number.')

        if PAHO_MQTT_VERSION >= (2, 0):
            return self._v2_reconnect(sock)
        if PAHO_MQTT_VERSION < (1, 6):
            # Paho Mqtt Version < 1.6.x
            self._in_packet = {
                "command": 0,
                "have_remaining": 0,
                "remaining_count": [],
                "remaining_mult": 1,
                "remaining_length": 0,
                "packet": b"",
                "to_process": 0,
                "pos": 0
            }
            with self._out_packet_mutex:
                self._out_packet = deque()  # type: ignore

            with self._current_out_packet_mutex:
                self._current_out_packet = None
        else:
            self._in_packet = {
                "command": 0,
                "have_remaining": 0,
                "remaining_count": [],
                "remaining_mult": 1,
                "remaining_length": 0,
                "packet": bytearray(b""),
                "to_process": 0,
                "pos": 0
            }
            self._out_packet = deque()  # type: ignore

        with self._msgtime_mutex:
            self._last_msg_in = paho_mqtt.time_func()
            self._last_msg_out = paho_mqtt.time_func()

        self._ping_t = 0
        self._state = paho_mqtt.mqtt_cs_new

        self._sock_close()

        # Put messages in progress in a valid state.
        self._messages_reconnect_reset()

        if sock is None:
            sock = self._create_socket_connection()

        if self._ssl:
            # SSL is only supported when SSLContext is available
            # (implies Python >= 2.7.9 or >= 3.2)

            verify_host = not self._tls_insecure
            try:
                # Try with server_hostname, even it's not supported in
                # certain scenarios
                sock = self._ssl_context.wrap_socket(
                    sock,
                    server_hostname=self._host,
                    do_handshake_on_connect=False,
                )
            except ssl.CertificateError:
                # CertificateError is derived from ValueError
                raise
            except ValueError:
                # Python version requires SNI in order to handle
                # server_hostname, but SNI is not available
                sock = self._ssl_context.wrap_socket(
                    sock,
                    do_handshake_on_connect=False,
                )
            else:
                # If SSL context has already checked hostname, then don't need
                #  to do it again
                if (hasattr(self._ssl_context, 'check_hostname') and
                        self._ssl_context.check_hostname):
                    verify_host = False

            assert isinstance(sock, ssl.SSLSocket)
            sock.settimeout(self._keepalive)
            sock.do_handshake()

            if verify_host:
                ssl.match_hostname(sock.getpeercert(), self._host)  # type: ignore

        if self._transport == "websockets":
            sock.settimeout(self._keepalive)
            sock = paho_mqtt.WebsocketWrapper(
                sock, self._host, self._port, self._ssl,
                self._websocket_path, self._websocket_extra_headers
            )

        self._sock = sock
        assert self._sock is not None
        self._sock.setblocking(False)
        self._registered_write = False
        self._call_socket_open()

        return self._send_connect(self._keepalive)

    def _v2_reconnect(self, sock: Optional[socket.socket] = None):
        self._in_packet = {
            "command": 0,
            "have_remaining": 0,
            "remaining_count": [],
            "remaining_mult": 1,
            "remaining_length": 0,
            "packet": bytearray(b""),
            "to_process": 0,
            "pos": 0,
        }

        self._ping_t = 0.0  # type: ignore
        self._state = paho_mqtt._ConnectionState.MQTT_CS_CONNECTING

        self._sock_close()

        # Mark all currently outgoing QoS = 0 packets as lost,
        # or `wait_for_publish()` could hang forever
        for pkt in self._out_packet:
            if (
                pkt["command"] & 0xF0 == paho_mqtt.PUBLISH and
                pkt["qos"] == 0 and pkt["info"] is not None
            ):
                pkt["info"].rc = paho_mqtt.MQTT_ERR_CONN_LOST
                pkt["info"]._set_as_published()

        self._out_packet.clear()

        with self._msgtime_mutex:
            self._last_msg_in = paho_mqtt.time_func()
            self._last_msg_out = paho_mqtt.time_func()

        # Put messages in progress in a valid state.
        self._messages_reconnect_reset()

        with self._callback_mutex:
            on_pre_connect = self.on_pre_connect

        if on_pre_connect:
            try:
                on_pre_connect(self, self._userdata)
            except Exception as err:
                self._easy_log(
                    paho_mqtt.MQTT_LOG_ERR,
                    'Caught exception in on_pre_connect: %s', err
                )
                if not self.suppress_exceptions:
                    raise

        self._sock = sock or self._create_socket()

        self._sock.setblocking(False)  # type: ignore[attr-defined]
        self._registered_write = False
        self._call_socket_open(self._sock)

        return self._send_connect(self._keepalive)


class SubscriptionHandle:
    def __init__(self, topic: str, callback: FlexCallback) -> None:
        self.callback = callback
        self.topic = topic

class BrokerAckLogger:
    def __init__(self, topics: List[str], action: str) -> None:
        self.topics = topics
        self.action = action

    def __call__(self, fut: asyncio.Future) -> None:
        if self.action == "subscribe":
            res: Union[List[int], List[paho_mqtt.ReasonCodes]]
            res = fut.result()
            log_msg = "MQTT Subscriptions Acknowledged"
            if len(res) != len(self.topics):
                log_msg += "\nTopic / QOS count mismatch, " \
                    f"\nTopics: {self.topics} " \
                    f"\nQoS responses: {res}"
            else:
                for topic, qos in zip(self.topics, res):
                    log_msg += f"\n Topic: {topic} | "
                    if isinstance(qos, paho_mqtt.ReasonCodes):
                        log_msg += qos.getName()
                    else:
                        log_msg += f"Granted QoS {qos}"
        elif self.action == "unsubscribe":
            log_msg = "MQTT Unsubscribe Acknowledged"
            for topic in self.topics:
                log_msg += f"\n Topic: {topic}"
        else:
            log_msg = f"Unknown action: {self.action}"
        logging.debug(log_msg)


SubscribedDict = Dict[str, Tuple[int, List[SubscriptionHandle]]]

class AIOHelper:
    def __init__(self, client: paho_mqtt.Client) -> None:
        self.loop = asyncio.get_running_loop()
        self.client = client
        self.client.on_socket_open = self._on_socket_open
        self.client.on_socket_close = self._on_socket_close
        self.client._on_socket_register_write = self._on_socket_register_write
        self.client._on_socket_unregister_write = \
            self._on_socket_unregister_write
        self.misc_task: Optional[asyncio.Task] = None

    def _on_socket_open(self,
                        client: paho_mqtt.Client,
                        userdata: Any,
                        sock: socket.socket
                        ) -> None:
        logging.info("MQTT Socket Opened")
        self.loop.add_reader(sock, client.loop_read)
        self.misc_task = self.loop.create_task(self.misc_loop())

    def _on_socket_close(self,
                         client: paho_mqtt.Client,
                         userdata: Any,
                         sock: socket.socket
                         ) -> None:
        logging.info("MQTT Socket Closed")
        self.loop.remove_reader(sock)
        if self.misc_task is not None:
            self.misc_task.cancel()

    def _on_socket_register_write(self,
                                  client: paho_mqtt.Client,
                                  userdata: Any,
                                  sock: socket.socket
                                  ) -> None:
        self.loop.add_writer(sock, client.loop_write)

    def _on_socket_unregister_write(self,
                                    client: paho_mqtt.Client,
                                    userdata: Any,
                                    sock: socket.socket
                                    ) -> None:
        self.loop.remove_writer(sock)

    async def misc_loop(self) -> None:
        while self.client.loop_misc() == paho_mqtt.MQTT_ERR_SUCCESS:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break
        logging.info("MQTT Misc Loop Complete")


class MQTTClient(APITransport):
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.eventloop = self.server.get_event_loop()
        self.address: str = config.get('address')
        self.port: int = config.getint('port', 1883)
        self.tls_enabled: bool = config.getboolean("enable_tls", False)
        user = config.gettemplate('username', None)
        self.user_name: Optional[str] = None
        if user:
            self.user_name = user.render()
        pw_file_path = config.get('password_file', None, deprecate=True)
        pw_template = config.gettemplate('password', None)
        self.password: Optional[str] = None
        if pw_file_path is not None:
            pw_file = pathlib.Path(pw_file_path).expanduser().absolute()
            if not pw_file.exists():
                raise config.error(
                    f"Password file '{pw_file}' does not exist")
            self.password = pw_file.read_text().strip()
        if pw_template is not None:
            self.password = pw_template.render()
        protocol = config.get('mqtt_protocol', "v3.1.1")
        self.protocol = MQTT_PROTOCOLS.get(protocol, None)
        if self.protocol is None:
            raise config.error(
                f"Invalid value '{protocol}' for option 'mqtt_protocol' "
                "in section [mqtt]. Must be one of "
                f"{MQTT_PROTOCOLS.values()}")
        self.instance_name = config.get('instance_name', socket.gethostname())
        if '+' in self.instance_name or '#' in self.instance_name:
            raise config.error(
                "Option 'instance_name' in section [mqtt] cannot "
                "contain a wildcard.")
        self.qos = config.getint("default_qos", 0)
        if self.qos > 2 or self.qos < 0:
            raise config.error(
                "Option 'default_qos' in section [mqtt] must be "
                "between 0 and 2")
        self.publish_split_status = \
            config.getboolean("publish_split_status", False)
        client_id: Optional[str] = config.get("client_id", None)
        if PAHO_MQTT_VERSION < (2, 0):
            self.client = ExtPahoClient(client_id, protocol=self.protocol)
        else:
            self.client = ExtPahoClient(
                paho_mqtt.CallbackAPIVersion.VERSION1, client_id,
                protocol=self.protocol
            )
        self.client.on_connect = self._on_connect
        self.client.on_message = self._on_message
        self.client.on_disconnect = self._on_disconnect
        self.client.on_publish = self._on_publish
        self.client.on_subscribe = self._on_subscribe
        self.client.on_unsubscribe = self._on_unsubscribe
        self.connect_evt: asyncio.Event = asyncio.Event()
        self.disconnect_evt: Optional[asyncio.Event] = None
        self.connect_task: Optional[asyncio.Task] = None
        self.subscribed_topics: SubscribedDict = {}
        self.pending_responses: List[asyncio.Future] = []
        self.pending_acks: Dict[int, asyncio.Future] = {}

        # We don't need to register these endpoints over the MQTT transport as they
        # are redundant.  MQTT clients can already publish and subscribe.
        ep_transports = TransportType.all() & ~TransportType.MQTT
        self.server.register_endpoint(
            "/server/mqtt/publish", RequestType.POST, self._handle_publish_request,
            transports=ep_transports
        )
        self.server.register_endpoint(
            "/server/mqtt/subscribe", RequestType.POST,
            self._handle_subscription_request,
            transports=ep_transports
        )

        # Subscribe to API requests
        self.api_request_topic = f"{self.instance_name}/moonraker/api/request"
        self.api_resp_topic = f"{self.instance_name}/moonraker/api/response"
        self.klipper_status_topic = f"{self.instance_name}/klipper/status"
        self.klipper_state_prefix = f"{self.instance_name}/klipper/state"
        self.moonraker_status_topic = f"{self.instance_name}/moonraker/status"
        status_cfg: Dict[str, str] = config.getdict(
            "status_objects", {}, allow_empty_fields=True
        )
        self.status_interval = config.getfloat("status_interval", 0, above=.25)
        self.status_cache: Dict[str, Dict[str, Any]] = {}
        self.status_update_timer: Optional[FlexTimer] = None
        self.last_status_time = 0.
        self.status_objs: Dict[str, Optional[List[str]]] = {}
        for key, val in status_cfg.items():
            if val is not None:
                self.status_objs[key] = [v.strip() for v in val.split(',') if v.strip()]
            else:
                self.status_objs[key] = None
        if status_cfg:
            logging.debug(f"MQTT: Status Objects Set: {self.status_objs}")
            self.server.register_event_handler(
                "server:klippy_started", self._handle_klippy_started
            )
            self.server.register_event_handler(
                "server:klippy_disconnect", self._handle_klippy_disconnect
            )
            if self.status_interval:
                self.status_update_timer = self.eventloop.register_timer(
                    self._handle_timed_status_update
                )

        self.timestamp_deque: Deque = deque(maxlen=20)
        self.api_qos = config.getint('api_qos', self.qos)
        if config.getboolean("enable_moonraker_api", True):
            self.subscribe_topic(self.api_request_topic,
                                 self._process_api_request,
                                 self.api_qos)

        self.server.register_remote_method("publish_mqtt_topic",
                                           self._publish_from_klipper)
        logging.info(
            f"\nReserved MQTT topics:\n"
            f"API Request: {self.api_request_topic}\n"
            f"API Response: {self.api_resp_topic}\n"
            f"Moonraker Status: {self.moonraker_status_topic}\n"
            f"Klipper Status: {self.klipper_status_topic}")

    async def component_init(self) -> None:
        # We must wait for the IOLoop (asyncio event loop) to start
        # prior to retrieving it
        self.helper = AIOHelper(self.client)
        if self.user_name is not None:
            self.client.username_pw_set(self.user_name, self.password)
        self.client.will_set(self.moonraker_status_topic,
                             payload=jsonw.dumps({'server': 'offline'}),
                             qos=self.qos, retain=True)
        if self.tls_enabled:
            self.client.tls_set()
        self.client.connect_async(self.address, self.port)
        self.connect_task = self.eventloop.create_task(
            self._do_reconnect(first=True)
        )

    async def _handle_klippy_started(self, state: KlippyState) -> None:
        if self.status_objs:
            kapi: KlippyAPI = self.server.lookup_component("klippy_apis")
            await kapi.subscribe_from_transport(
                self.status_objs, self, default=None,
            )
            if self.status_update_timer is not None:
                self.status_update_timer.start(delay=self.status_interval)

    def _handle_klippy_disconnect(self):
        if self.status_update_timer is not None:
            self.status_update_timer.stop()
        if self.status_cache:
            payload = self.status_cache
            self.status_cache = {}
            self._publish_status_update(payload, self.last_status_time)

    def _on_message(self,
                    client: str,
                    user_data: Any,
                    message: paho_mqtt.MQTTMessage
                    ) -> None:
        topic = message.topic
        if topic in self.subscribed_topics:
            cb_hdls = self.subscribed_topics[topic][1]
            for hdl in cb_hdls:
                self.eventloop.register_callback(
                    hdl.callback, message.payload)
        else:
            logging.debug(
                f"Unregistered MQTT Topic Received: {topic}, "
                f"payload: {message.payload.decode()}")

    def _on_connect(self,
                    client: paho_mqtt.Client,
                    user_data: Any,
                    flags: Dict[str, Any],
                    reason_code: Union[int, paho_mqtt.ReasonCodes],
                    properties: Optional[paho_mqtt.Properties] = None
                    ) -> None:
        logging.info("MQTT Client Connected")
        if reason_code == 0:
            self.publish_topic(self.moonraker_status_topic,
                               {'server': 'online'}, retain=True)
            subs = [(k, v[0]) for k, v in self.subscribed_topics.items()]
            if subs:
                res, msg_id = client.subscribe(subs)
                if msg_id is not None:
                    sub_fut: asyncio.Future = self.eventloop.create_future()
                    topics = list(self.subscribed_topics.keys())
                    sub_fut.add_done_callback(
                        BrokerAckLogger(topics, "subscribe"))
                    self.pending_acks[msg_id] = sub_fut
            self.connect_evt.set()
            self.server.send_event("mqtt:connected")
        else:
            if isinstance(reason_code, int):
                err_str = paho_mqtt.connack_string(reason_code)
            else:
                err_str = reason_code.getName()
            self.server.set_failed_component("mqtt")
            self.server.add_warning(f"MQTT Connection Failed: {err_str}")

    def _on_disconnect(self,
                       client: paho_mqtt.Client,
                       user_data: Any,
                       reason_code: int,
                       properties: Optional[paho_mqtt.Properties] = None
                       ) -> None:
        if self.disconnect_evt is not None:
            self.disconnect_evt.set()
        elif self.is_connected():
            # The server connection was dropped, attempt to reconnect
            logging.info("MQTT Server Disconnected, reason: "
                         f"{paho_mqtt.error_string(reason_code)}")
            if self.connect_task is None:
                self.connect_task = asyncio.create_task(self._do_reconnect())
            self.server.send_event("mqtt:disconnected")
        self.connect_evt.clear()

    def _on_publish(self,
                    client: paho_mqtt.Client,
                    user_data: Any,
                    msg_id: int
                    ) -> None:
        pub_fut = self.pending_acks.pop(msg_id, None)
        if pub_fut is not None and not pub_fut.done():
            pub_fut.set_result(None)

    def _on_subscribe(self,
                      client: paho_mqtt.Client,
                      user_data: Any,
                      msg_id: int,
                      flex: Union[List[int], List[paho_mqtt.ReasonCodes]],
                      properties: Optional[paho_mqtt.Properties] = None
                      ) -> None:
        sub_fut = self.pending_acks.pop(msg_id, None)
        if sub_fut is not None and not sub_fut.done():
            sub_fut.set_result(flex)

    def _on_unsubscribe(self,
                        client: paho_mqtt.Client,
                        user_data: Any,
                        msg_id: int,
                        properties: Optional[paho_mqtt.Properties] = None,
                        reasoncodes: Optional[paho_mqtt.ReasonCodes] = None
                        ) -> None:
        unsub_fut = self.pending_acks.pop(msg_id, None)
        if unsub_fut is not None and not unsub_fut.done():
            unsub_fut.set_result(None)

    async def _do_reconnect(self, first: bool = False) -> None:
        logging.info("Attempting MQTT Connect/Reconnect")
        last_err: Exception = Exception()
        while True:
            if not first:
                try:
                    await asyncio.sleep(2.)
                except asyncio.CancelledError:
                    raise
            first = False
            try:
                sock = await self.eventloop.create_socket_connection(
                    (self.address, self.port), timeout=10
                )
                self.client.reconnect(sock)
            except asyncio.CancelledError:
                raise
            except Exception as e:
                if type(last_err) is not type(e) or last_err.args != e.args:
                    logging.exception("MQTT Connection Error")
                    last_err = e
                continue
            self.client.socket().setsockopt(
                socket.SOL_SOCKET, socket.SO_SNDBUF, 2048)
            break
        self.connect_task = None

    async def wait_connection(self, timeout: Optional[float] = None) -> bool:
        try:
            await asyncio.wait_for(self.connect_evt.wait(), timeout)
        except asyncio.TimeoutError:
            return False
        return True

    def is_connected(self) -> bool:
        return self.connect_evt.is_set()

    def subscribe_topic(self,
                        topic: str,
                        callback: FlexCallback,
                        qos: Optional[int] = None
                        ) -> SubscriptionHandle:
        if '#' in topic or '+' in topic:
            raise self.server.error("Wildcards may not be used")
        qos = qos or self.qos
        if qos > 2 or qos < 0:
            raise self.server.error("QOS must be between 0 and 2")
        hdl = SubscriptionHandle(topic, callback)
        sub_handles = [hdl]
        need_sub = True
        if topic in self.subscribed_topics:
            prev_qos, sub_handles = self.subscribed_topics[topic]
            qos = max(qos, prev_qos)
            sub_handles.append(hdl)
            need_sub = qos != prev_qos
        self.subscribed_topics[topic] = (qos, sub_handles)
        if self.is_connected() and need_sub:
            res, msg_id = self.client.subscribe(topic, qos)
            if msg_id is not None:
                sub_fut: asyncio.Future = self.eventloop.create_future()
                sub_fut.add_done_callback(
                    BrokerAckLogger([topic], "subscribe"))
                self.pending_acks[msg_id] = sub_fut
        return hdl

    def unsubscribe(self, hdl: SubscriptionHandle) -> None:
        topic = hdl.topic
        if topic in self.subscribed_topics:
            sub_hdls = self.subscribed_topics[topic][1]
            try:
                sub_hdls.remove(hdl)
            except Exception:
                pass
            if not sub_hdls:
                del self.subscribed_topics[topic]
                res, msg_id = self.client.unsubscribe(topic)
                if msg_id is not None:
                    unsub_fut: asyncio.Future = self.eventloop.create_future()
                    unsub_fut.add_done_callback(
                        BrokerAckLogger([topic], "unsubscribe"))
                    self.pending_acks[msg_id] = unsub_fut

    def publish_topic(self,
                      topic: str,
                      payload: Any = None,
                      qos: Optional[int] = None,
                      retain: bool = False
                      ) -> Awaitable[None]:
        qos = qos or self.qos
        if qos > 2 or qos < 0:
            raise self.server.error("QOS must be between 0 and 2")
        pub_fut: asyncio.Future = self.eventloop.create_future()
        if isinstance(payload, (dict, list)):
            try:
                payload = jsonw.dumps(payload)
            except jsonw.JSONDecodeError:
                raise self.server.error(
                    "Dict or List is not json encodable") from None
        elif isinstance(payload, bool):
            payload = str(payload).lower()
        try:
            msg_info = self.client.publish(topic, payload, qos, retain)
            if msg_info.is_published():
                pub_fut.set_result(None)
            else:
                if qos == 0:
                    # There is no delivery guarantee for qos == 0, so
                    # it is possible that the on_publish event will
                    # not be called if paho mqtt encounters an error
                    # during publication.  Return immediately as
                    # a workaround.
                    if msg_info.rc != paho_mqtt.MQTT_ERR_SUCCESS:
                        err_str = paho_mqtt.error_string(msg_info.rc)
                        pub_fut.set_exception(self.server.error(
                            f"MQTT Publish Error: {err_str}", 503))
                    else:
                        pub_fut.set_result(None)
                    return pub_fut
                self.pending_acks[msg_info.mid] = pub_fut
        except ValueError:
            pub_fut.set_exception(self.server.error(
                "MQTT Message Queue Full", 529))
        except Exception as e:
            pub_fut.set_exception(self.server.error(
                f"MQTT Publish Error: {e}", 503))
        return pub_fut

    async def publish_topic_with_response(self,
                                          topic: str,
                                          response_topic: str,
                                          payload: Any = None,
                                          qos: Optional[int] = None,
                                          retain: bool = False,
                                          timeout: Optional[float] = None
                                          ) -> bytes:
        qos = qos or self.qos
        if qos > 2 or qos < 0:
            raise self.server.error("QOS must be between 0 and 2")
        resp_fut: asyncio.Future = self.eventloop.create_future()
        resp_hdl = self.subscribe_topic(
            response_topic, resp_fut.set_result, qos)
        self.pending_responses.append(resp_fut)
        try:
            await asyncio.wait_for(self.publish_topic(
                topic, payload, qos, retain), timeout)
            await asyncio.wait_for(resp_fut, timeout)
        except asyncio.TimeoutError:
            logging.info(f"Response to request {topic} timed out")
            raise self.server.error("MQTT Request Timed Out", 504)
        finally:
            try:
                self.pending_responses.remove(resp_fut)
            except Exception:
                pass
            self.unsubscribe(resp_hdl)
        return resp_fut.result()

    async def _handle_publish_request(self,
                                      web_request: WebRequest
                                      ) -> Dict[str, Any]:
        topic: str = web_request.get_str("topic")
        payload: Any = web_request.get("payload", None)
        qos: int = web_request.get_int("qos", self.qos)
        retain: bool = web_request.get_boolean("retain", False)
        timeout: Optional[float] = web_request.get_float('timeout', None)
        try:
            await asyncio.wait_for(self.publish_topic(
                topic, payload, qos, retain), timeout)
        except asyncio.TimeoutError:
            raise self.server.error("MQTT Publish Timed Out", 504)
        return {
            "topic": topic
        }

    async def _handle_subscription_request(self,
                                           web_request: WebRequest
                                           ) -> Dict[str, Any]:
        topic: str = web_request.get_str("topic")
        qos: int = web_request.get_int("qos", self.qos)
        timeout: Optional[float] = web_request.get_float('timeout', None)
        resp: asyncio.Future = self.eventloop.create_future()
        hdl: Optional[SubscriptionHandle] = None
        try:
            hdl = self.subscribe_topic(topic, resp.set_result, qos)
            self.pending_responses.append(resp)
            await asyncio.wait_for(resp, timeout)
            ret: bytes = resp.result()
        except asyncio.TimeoutError:
            raise self.server.error("MQTT Subscribe Timed Out", 504)
        finally:
            try:
                self.pending_responses.remove(resp)
            except Exception:
                pass
            if hdl is not None:
                self.unsubscribe(hdl)
        try:
            payload = jsonw.loads(ret)
        except jsonw.JSONDecodeError:
            payload = ret.decode()
        return {
            'topic': topic,
            'payload': payload
        }

    async def _process_api_request(self, payload: bytes) -> None:
        rpc: JsonRPC = self.server.lookup_component("jsonrpc")
        response = await rpc.dispatch(payload, self)
        if response is not None:
            await self.publish_topic(self.api_resp_topic, response,
                                     self.api_qos)

    @property
    def transport_type(self) -> TransportType:
        return TransportType.MQTT

    def screen_rpc_request(
        self, api_def: APIDefinition, req_type: RequestType, args: Dict[str, Any]
    ) -> None:
        ts = args.pop("mqtt_timestamp", None)
        if ts is not None:
            if ts in self.timestamp_deque:
                logging.debug("Duplicate MQTT API request received")
                raise self.server.error(
                    "Duplicate MQTT Request", DUP_API_REQ_CODE)
            else:
                self.timestamp_deque.append(ts)

    def send_status(self, status: Dict[str, Any], eventtime: float) -> None:
        if not status or not self.is_connected():
            return
        if not self.status_interval:
            self._publish_status_update(status, eventtime)
        else:
            for key, val in status.items():
                self.status_cache.setdefault(key, {}).update(val)
            self.last_status_time = eventtime

    def _handle_timed_status_update(self, eventtime: float) -> float:
        if self.status_cache:
            payload = self.status_cache
            self.status_cache = {}
            self._publish_status_update(payload, self.last_status_time)
        return eventtime + self.status_interval

    def _publish_status_update(self, status: Dict[str, Any], eventtime: float) -> None:
        if self.publish_split_status:
            for objkey in status:
                objval = status[objkey]
                for statekey in objval:
                    payload = {'eventtime': eventtime, 'value': objval[statekey]}
                    self.publish_topic(
                        f"{self.klipper_state_prefix}/{objkey}/{statekey}",
                        payload, retain=True)
        else:
            payload = {'eventtime': eventtime, 'status': status}
            self.publish_topic(self.klipper_status_topic, payload)


    def get_instance_name(self) -> str:
        return self.instance_name

    async def close(self) -> None:
        if self.status_update_timer is not None:
            self.status_update_timer.stop()
        if self.connect_task is not None:
            self.connect_task.cancel()
            self.connect_task = None
        if not self.is_connected():
            return
        await self.publish_topic(self.moonraker_status_topic,
                                 {'server': 'offline'},
                                 retain=True)
        self.disconnect_evt = asyncio.Event()
        self.client.disconnect()
        try:
            await asyncio.wait_for(self.disconnect_evt.wait(), 2.)
        except asyncio.TimeoutError:
            logging.info("MQTT Disconnect Timeout")
        futs = list(self.pending_acks.values())
        futs.extend(self.pending_responses)
        for fut in futs:
            if fut.done():
                continue
            fut.set_exception(
                self.server.error("Moonraker Shutdown", 503))

    async def _publish_from_klipper(self,
                                    topic: str,
                                    payload: Any = None,
                                    qos: Optional[int] = None,
                                    retain: bool = False,
                                    use_prefix: bool = False
                                    ) -> None:
        if use_prefix:
            topic = f"{self.instance_name}/{topic.lstrip('/')}"
        await self.publish_topic(topic, payload, qos, retain)


def load_component(config: ConfigHelper) -> MQTTClient:
    return MQTTClient(config)



from __future__ import annotations
import pytest
import pytest_asyncio
import asyncio
import shutil
import re
import pathlib
import sys
import shlex
import tempfile
import subprocess
from typing import Iterator, Dict, AsyncIterator, Any
from moonraker.server import Server
from moonraker.eventloop import EventLoop
from moonraker import utils
import dbtool
from fixtures import KlippyProcess, HttpClient, WebsocketClient

ASSETS = pathlib.Path(__file__).parent.joinpath("assets")

need_klippy_restart = pytest.StashKey[bool]()

def pytest_addoption(parser: pytest.Parser, pluginmanager):
    parser.addoption("--klipper-path", action="store", dest="klipper_path")
    parser.addoption("--klipper-exec", action="store", dest="klipper_exec")

def interpolate_config(source_path: pathlib.Path,
                       dest_path: pathlib.Path,
                       keys: Dict[str, Any]
                       ) -> None:
    def interp(match):
        return str(keys[match.group(1)])
    sub_data = re.sub(r"\${([^}]+)}", interp, source_path.read_text())
    dest_path.write_text(sub_data)

@pytest.fixture(scope="session", autouse=True)
def ssl_certs() -> Iterator[Dict[str, pathlib.Path]]:
    with tempfile.TemporaryDirectory(prefix="moonraker-certs-") as tmpdir:
        tmp_path = pathlib.Path(tmpdir)
        cert_path = tmp_path.joinpath("certificate.pem")
        key_path = tmp_path.joinpath("privkey.pem")
        cmd = (
            f"openssl req -newkey rsa:4096 -nodes -keyout {key_path} "
            f"-x509 -days 365 -out {cert_path} -sha256 "
            "-subj '/C=US/ST=NRW/L=Earth/O=Moonraker/OU=IT/"
            "CN=www.moonraker-test.com/emailAddress=mail@moonraker-test.com'"
        )
        args = shlex.split(cmd)
        subprocess.run(args, check=True)
        yield {
            "ssl_certificate_path": cert_path,
            "ssl_key_path": key_path,
        }

@pytest.fixture(scope="class")
def event_loop() -> Iterator[asyncio.AbstractEventLoop]:
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
def session_args(ssl_certs: Dict[str, pathlib.Path]
                 ) -> Iterator[Dict[str, pathlib.Path]]:
    mconf_asset = ASSETS.joinpath(f"moonraker/base_server.conf")
    secrets_asset = ASSETS.joinpath(f"moonraker/secrets.ini")
    pcfg_asset = ASSETS.joinpath(f"klipper/base_printer.cfg")
    with tempfile.TemporaryDirectory(prefix="moonraker-test") as tmpdir:
        tmp_path = pathlib.Path(tmpdir)
        secrets_dest = tmp_path.joinpath("secrets.ini")
        shutil.copy(secrets_asset, secrets_dest)
        cfg_path = tmp_path.joinpath("config")
        cfg_path.mkdir()
        log_path = tmp_path.joinpath("logs")
        log_path.mkdir()
        db_path = tmp_path.joinpath("database")
        db_path.mkdir()
        gcode_path = tmp_path.joinpath("gcode_files")
        gcode_path.mkdir()
        dest_paths = {
            "temp_path": tmp_path,
            "asset_path": ASSETS,
            "config_path": cfg_path,
            "database_path": db_path,
            "log_path": log_path,
            "gcode_path": gcode_path,
            "secrets_path": secrets_dest,
            "klippy_uds_path": tmp_path.joinpath("klippy_uds"),
            "klippy_pty_path": tmp_path.joinpath("klippy_pty"),
            "klipper.dict": ASSETS.joinpath("klipper/klipper.dict"),
            "mconf_asset": mconf_asset,
            "pcfg_asset": pcfg_asset,
        }
        dest_paths.update(ssl_certs)
        mconf_dest = cfg_path.joinpath("moonraker.conf")
        dest_paths["moonraker.conf"] = mconf_dest
        interpolate_config(mconf_asset, mconf_dest, dest_paths)
        pcfg_dest = cfg_path.joinpath("printer.cfg")
        dest_paths["printer.cfg"] = pcfg_dest
        interpolate_config(pcfg_asset, pcfg_dest, dest_paths)
        yield dest_paths

@pytest.fixture(scope="session")
def klippy_session(session_args: Dict[str, pathlib.Path],
                   pytestconfig: pytest.Config) -> Iterator[KlippyProcess]:
    pytestconfig.stash[need_klippy_restart] = False
    kpath = pytestconfig.getoption('klipper_path', "~/klipper")
    kexec = pytestconfig.getoption('klipper_exec', None)
    if kexec is None:
        kexec = sys.executable
    exec = pathlib.Path(kexec).expanduser()
    klipper_path = pathlib.Path(kpath).expanduser()
    base_cmd = f"{exec} {klipper_path}/klippy/klippy.py "
    kproc = KlippyProcess(base_cmd, session_args)
    kproc.start()
    yield kproc
    kproc.stop()

@pytest.fixture(scope="class")
def klippy(klippy_session: KlippyProcess,
           pytestconfig: pytest.Config):
    if pytestconfig.stash[need_klippy_restart]:
        pytestconfig.stash[need_klippy_restart] = False
        klippy_session.restart()
    return klippy_session

@pytest.fixture(scope="class")
def path_args(request: pytest.FixtureRequest,
              session_args: Dict[str, pathlib.Path],
              pytestconfig: pytest.Config
              ) -> Iterator[Dict[str, pathlib.Path]]:
    path_marker = request.node.get_closest_marker("run_paths")
    paths: Dict[str, Any] = {
        "moonraker_conf": "base_server.conf",
        "secrets": "secrets.ini",
        "printer_cfg": "base_printer.cfg",
        "klippy_uds": None,
    }
    if path_marker is not None:
        paths.update(path_marker.kwargs)
    tmp_path = session_args["temp_path"]
    cfg_path = session_args["config_path"]
    mconf_dest = session_args["moonraker.conf"]
    mconf_asset = ASSETS.joinpath(f"moonraker/{paths['moonraker_conf']}")
    pcfg_asset = ASSETS.joinpath(f"klipper/{paths['printer_cfg']}")
    last_uds = session_args["klippy_uds_path"]
    if paths["klippy_uds"] is not None:
        tmp_uds = tmp_path.joinpath(paths["klippy_uds"])
        session_args["klippy_uds_path"] = tmp_uds
    if (
        not mconf_asset.samefile(session_args["mconf_asset"]) or
        paths["klippy_uds"] is not None
    ):
        session_args['mconf_asset'] = mconf_asset
        interpolate_config(mconf_asset, mconf_dest, session_args)
    if not pcfg_asset.samefile(session_args["pcfg_asset"]):
        pcfg_dest = session_args["printer.cfg"]
        session_args["pcfg_asset"] = pcfg_asset
        interpolate_config(pcfg_asset, pcfg_dest, session_args)
        pytestconfig.stash[need_klippy_restart] = True
    if paths["secrets"] != session_args["secrets_path"].name:
        secrets_asset = ASSETS.joinpath(f"moonraker/{paths['secrets']}")
        secrets_dest = tmp_path.joinpath(paths['secrets'])
        shutil.copy(secrets_asset, secrets_dest)
        session_args["secrets_path"] = secrets_dest
    if "moonraker_log" in paths:
        log_path = session_args["log_path"]
        session_args['moonraker.log'] = log_path.joinpath(
            paths["moonraker_log"])
    bkp_dest: pathlib.Path = cfg_path.joinpath(".moonraker.conf.bkp")
    if "moonraker_bkp" in paths:
        bkp_source = ASSETS.joinpath("moonraker/base_server.conf")
        bkp_dest = cfg_path.joinpath(paths["moonraker_bkp"])
        interpolate_config(bkp_source, bkp_dest, session_args)
    if "database" in paths:
        db_source = ASSETS.joinpath(f"moonraker/{paths['database']}")
        db_dest = session_args["database_path"]
        db_args = {"input": str(db_source), "destination": db_dest}
        dbtool.restore(db_args)
    yield session_args
    log = session_args.pop("moonraker.log", None)
    if log is not None and log.is_file():
        log.unlink()
    if bkp_dest.is_file():
        bkp_dest.unlink()
    for item in session_args["database_path"].iterdir():
        if item.is_file():
            item.unlink()
    session_args["klippy_uds_path"] = last_uds
    if paths["klippy_uds"] is not None:
        # restore the original uds path
        interpolate_config(mconf_asset, mconf_dest, session_args)

@pytest.fixture(scope="class")
def base_server(path_args: Dict[str, pathlib.Path],
                event_loop: asyncio.AbstractEventLoop
                ) -> Iterator[Server]:
    evtloop = EventLoop()
    args = {
        'config_file': str(path_args['moonraker.conf']),
        'log_file': str(path_args.get("moonraker.log", "")),
        'software_version': "moonraker-pytest"
    }
    ql = logger = None
    if args["log_file"]:
        ql, logger, warning = utils.setup_logging(args)
        if warning:
            args["log_warning"] = warning
    yield Server(args, logger, evtloop)
    if ql is not None:
        ql.stop()

@pytest_asyncio.fixture(scope="class")
async def full_server(base_server: Server) -> AsyncIterator[Server]:
    base_server.load_components()
    ret = base_server.server_init(start_server=False)
    await asyncio.wait_for(ret, 4.)
    yield base_server
    if base_server.event_loop.aioloop.is_running():
        await base_server._stop_server(exit_reason="terminate")

@pytest_asyncio.fixture(scope="class")
async def ready_server(full_server: Server, klippy: KlippyProcess):
    ret = full_server.start_server(connect_to_klippy=False)
    await asyncio.wait_for(ret, 4.)
    ret = full_server.klippy_connection.connect()
    await asyncio.wait_for(ret, 4.)
    yield full_server

@pytest_asyncio.fixture(scope="class")
async def http_client() -> AsyncIterator[HttpClient]:
    client = HttpClient()
    yield client
    client.close()

@pytest_asyncio.fixture(scope="class")
async def websocket_client(request: pytest.FixtureRequest
                           ) -> AsyncIterator[WebsocketClient]:
    conn_marker = request.node.get_closest_marker("no_ws_connect")
    client = WebsocketClient()
    if conn_marker is None:
        await client.connect()
    yield client
    client.close()



#! /usr/bin/python3
# Create system dependencies json file from the install script
#
# Copyright (C) 2023 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license
from __future__ import annotations
import argparse
import pathlib
import json
import re
from typing import List, Dict

def make_sysdeps(input: str, output: str, distro: str, truncate: bool) -> None:
    sysdeps: Dict[str, List[str]] = {}
    outpath = pathlib.Path(output).expanduser().resolve()
    if outpath.is_file() and not truncate:
        sysdeps = json.loads(outpath.read_bytes())
    inst_path: pathlib.Path = pathlib.Path(input).expanduser().resolve()
    if not inst_path.is_file():
        raise Exception(f"Unable to locate install script: {inst_path}")
    data = inst_path.read_text()
    plines: List[str] = re.findall(r'PKGLIST="(.*)"', data)
    plines = [p.lstrip("${PKGLIST}").strip() for p in plines]
    packages: List[str] = []
    for line in plines:
        packages.extend(line.split())
    sysdeps[distro] = packages
    outpath.write_text(json.dumps(sysdeps, indent=4))


if __name__ == "__main__":
    def_path = pathlib.Path(__file__).parent
    desc = (
        "make_sysdeps - generate system dependency json file from an install script"
    )
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument(
        "-i", "--input", metavar="<install script>",
        help="path of the install script to read",
        default=f"{def_path}/install-moonraker.sh"
    )
    parser.add_argument(
        "-o", "--output", metavar="<output file>",
        help="path of the system dependency file to write",
        default=f"{def_path}/system-dependencies.json"
    )
    parser.add_argument(
        "-d", "--distro", metavar="<linux distro>",
        help="linux distro for dependencies", default="debian"
    )
    parser.add_argument(
        "-t", "--truncate", action="store_true",
        help="truncate output file"
    )
    args = parser.parse_args()
    make_sysdeps(args.input, args.output, args.distro, args.truncate)



# linux shell command execution utility
#
# Copyright (C) 2020  Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import os
import shlex
import logging
import signal
import asyncio
from ..utils import ServerError

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Awaitable,
    List,
    Optional,
    Callable,
    Coroutine,
    Dict,
    Set,
    cast
)
if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    OutputCallback = Optional[Callable[[bytes], None]]

class ShellCommandError(ServerError):
    def __init__(
        self,
        message: str,
        return_code: Optional[int],
        stdout: Optional[bytes] = b"",
        stderr: Optional[bytes] = b"",
        status_code: int = 500
    ) -> None:
        super().__init__(message, status_code=status_code)
        self.stdout = stdout or b""
        self.stderr = stderr or b""
        self.return_code = return_code

class ShellCommandProtocol(asyncio.subprocess.SubprocessStreamProtocol):
    def __init__(
        self,
        limit: int,
        loop: asyncio.events.AbstractEventLoop,
        std_out_cb: OutputCallback = None,
        std_err_cb: OutputCallback = None,
        log_stderr: bool = False
    ) -> None:
        self._loop = loop
        self._pipe_fds: List[int] = []
        super().__init__(limit, loop)
        self.std_out_cb = std_out_cb
        self.std_err_cb = std_err_cb
        self.log_stderr = log_stderr
        self.pending_data: List[bytes] = [b"", b""]

    def connection_made(
        self, transport: asyncio.transports.BaseTransport
    ) -> None:
        transport = cast(asyncio.SubprocessTransport, transport)
        self._transport = transport
        stdout_transport = transport.get_pipe_transport(1)
        if stdout_transport is not None:
            self._pipe_fds.append(1)

        stderr_transport = transport.get_pipe_transport(2)
        if stderr_transport is not None:
            self._pipe_fds.append(2)

        stdin_transport = transport.get_pipe_transport(0)
        if stdin_transport is not None:
            self.stdin = asyncio.streams.StreamWriter(
                stdin_transport,  # type: ignore
                protocol=self,
                reader=None,
                loop=self._loop
            )

    def pipe_data_received(self, fd: int, data: bytes | str) -> None:
        cb = None
        data_idx = fd - 1
        if fd == 1:
            cb = self.std_out_cb
        elif fd == 2:
            cb = self.std_err_cb
            if self.log_stderr:
                if isinstance(data, bytes):
                    msg = data.decode(errors='ignore')
                else:
                    msg = data
                logging.info(msg)
        if cb is not None:
            if isinstance(data, str):
                data = data.encode()
            lines = data.split(b'\n')
            lines[0] = self.pending_data[data_idx] + lines[0]
            self.pending_data[data_idx] = lines.pop()
            for line in lines:
                if not line:
                    continue
                cb(line)

    def pipe_connection_lost(
        self, fd: int, exc: Exception | None
    ) -> None:
        cb = None
        pending = b""
        if fd == 1:
            cb = self.std_out_cb
            pending = self.pending_data[0]
        elif fd == 2:
            cb = self.std_err_cb
            pending = self.pending_data[1]
        if pending and cb is not None:
            cb(pending)
        super().pipe_connection_lost(fd, exc)


class ShellCommand:
    IDX_SIGINT = 0
    IDX_SIGTERM = 1
    IDX_SIGKILL = 2
    def __init__(
        self,
        factory: ShellCommandFactory,
        cmd: str,
        std_out_callback: OutputCallback,
        std_err_callback: OutputCallback,
        env: Optional[Dict[str, str]] = None,
        log_stderr: bool = False,
        cwd: Optional[str] = None
    ) -> None:
        self.factory = factory
        self.name = cmd
        self.std_out_cb = std_out_callback
        self.std_err_cb = std_err_callback
        cmd = os.path.expanduser(cmd)
        self.command = shlex.split(cmd)
        self.log_stderr = log_stderr
        self.env = env
        self.cwd = cwd
        self.proc: Optional[asyncio.subprocess.Process] = None
        self.cancelled = False
        self.return_code: Optional[int] = None
        self.run_lock = asyncio.Lock()

    async def cancel(self, sig_idx: int = 1) -> None:
        if self.cancelled:
            return
        self.cancelled = True
        if self.proc is not None:
            exit_success = False
            sig_idx = min(2, max(0, sig_idx))
            sigs = [signal.SIGINT, signal.SIGTERM, signal.SIGKILL][sig_idx:]
            for sig in sigs:
                try:
                    self.proc.send_signal(sig)
                    ret = self.proc.wait()
                    await asyncio.wait_for(ret, timeout=2.)
                except asyncio.TimeoutError:
                    continue
                except ProcessLookupError:
                    pass
                logging.debug(f"Command '{self.name}' exited with "
                              f"signal: {sig.name}")
                exit_success = True
                break
            if not exit_success:
                logging.info(f"WARNING: {self.name} did not cleanly exit")

    def get_return_code(self) -> Optional[int]:
        return self.return_code

    def _reset_command_data(self) -> None:
        self.return_code = self.proc = None
        self.cancelled = False

    async def run(
        self,
        timeout: float = 2.,
        verbose: bool = True,
        log_complete: bool = True,
        sig_idx: int = 1,
        proc_input: Optional[str] = None,
        success_codes: Optional[List[int]] = None
    ) -> bool:
        async with self.run_lock:
            self.factory.add_running_command(self)
            self._reset_command_data()
            if not timeout:
                # Never timeout
                timeout = 9999999999999999.
            if (
                self.std_out_cb is None
                and self.std_err_cb is None and
                not self.log_stderr
            ):
                # No callbacks set so output cannot be verbose
                verbose = False
            created = await self._create_subprocess(
                verbose, proc_input is not None)
            if not created:
                self.factory.remove_running_command(self)
                return False
            assert self.proc is not None
            try:
                if proc_input is not None:
                    ret: Coroutine = self.proc.communicate(
                        input=proc_input.encode())
                else:
                    ret = self.proc.wait()
                await asyncio.wait_for(ret, timeout=timeout)
            except asyncio.TimeoutError:
                complete = False
                await self.cancel(sig_idx)
            else:
                complete = not self.cancelled
            self.factory.remove_running_command(self)
            return self._check_proc_success(
                complete, log_complete, success_codes
            )

    async def run_with_response(
        self,
        timeout: float = 2.,
        attempts: int = 1,
        log_complete: bool = True,
        sig_idx: int = 1,
        proc_input: Optional[str] = None,
        success_codes: Optional[List[int]] = None
    ) -> str:
        async with self.run_lock:
            self.factory.add_running_command(self)
            attempts = max(1, attempts)
            stdin: Optional[bytes] = None
            if proc_input is not None:
                stdin = proc_input.encode()
            while attempts > 0:
                self._reset_command_data()
                timed_out = False
                stdout = stderr = b""
                if await self._create_subprocess(has_input=stdin is not None):
                    assert self.proc is not None
                    try:
                        ret = self.proc.communicate(input=stdin)
                        stdout, stderr = await asyncio.wait_for(
                            ret, timeout=timeout)
                    except asyncio.TimeoutError:
                        complete = False
                        timed_out = True
                        await self.cancel(sig_idx)
                    else:
                        complete = not self.cancelled
                        if self.log_stderr and stderr:
                            logging.info(
                                f"{self.command[0]}: "
                                f"{stderr.decode(errors='ignore')}")
                    if self._check_proc_success(
                        complete, log_complete, success_codes
                    ):
                        self.factory.remove_running_command(self)
                        return stdout.decode(errors='ignore').rstrip("\n")
                    if stdout:
                        logging.debug(
                            f"Shell command '{self.name}' output:"
                            f"\n{stdout.decode(errors='ignore')}")
                    if self.cancelled and not timed_out:
                        break
                attempts -= 1
                await asyncio.sleep(.5)
            self.factory.remove_running_command(self)
            raise ShellCommandError(
                f"Error running shell command: '{self.name}'",
                self.return_code, stdout, stderr)

    async def _create_subprocess(
        self,
        use_callbacks: bool = False,
        has_input: bool = False
    ) -> bool:
        loop = asyncio.get_running_loop()

        def protocol_factory():
            return ShellCommandProtocol(
                limit=2**20, loop=loop, std_out_cb=self.std_out_cb,
                std_err_cb=self.std_err_cb, log_stderr=self.log_stderr
            )
        try:
            stdpipe: Optional[int] = None
            if has_input:
                stdpipe = asyncio.subprocess.PIPE
            if self.std_err_cb is not None or self.log_stderr:
                errpipe = asyncio.subprocess.PIPE
            else:
                errpipe = asyncio.subprocess.STDOUT
            if use_callbacks:
                transport, protocol = await loop.subprocess_exec(
                    protocol_factory, *self.command, stdin=stdpipe,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=errpipe, env=self.env, cwd=self.cwd)
                self.proc = asyncio.subprocess.Process(
                    transport, protocol, loop)
            else:
                self.proc = await asyncio.create_subprocess_exec(
                    *self.command, stdin=stdpipe,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=errpipe, env=self.env, cwd=self.cwd)
        except asyncio.CancelledError:
            raise
        except Exception:
            logging.exception(
                f"shell_command: Command ({self.name}) failed")
            return False
        return True

    def _check_proc_success(
        self,
        complete: bool,
        log_complete: bool,
        success_codes: Optional[List[int]] = None
    ) -> bool:
        assert self.proc is not None
        if success_codes is None:
            success_codes = [0]
        self.return_code = self.proc.returncode
        success = self.return_code in success_codes and complete
        if success:
            msg = f"Command ({self.name}) successfully finished"
        elif self.cancelled:
            msg = f"Command ({self.name}) cancelled"
        elif not complete:
            msg = f"Command ({self.name}) timed out"
        else:
            msg = f"Command ({self.name}) exited with return code" \
                f" {self.return_code}"
        if log_complete:
            logging.info(msg)
        return success

class ShellCommandFactory:
    error = ShellCommandError
    def __init__(self, config: ConfigHelper) -> None:
        self.running_commands: Set[ShellCommand] = set()

    def add_running_command(self, cmd: ShellCommand) -> None:
        self.running_commands.add(cmd)

    def remove_running_command(self, cmd: ShellCommand) -> None:
        try:
            self.running_commands.remove(cmd)
        except KeyError:
            pass

    def build_shell_command(
        self,
        cmd: str,
        callback: OutputCallback = None,
        std_err_callback: OutputCallback = None,
        env: Optional[Dict[str, str]] = None,
        log_stderr: bool = False,
        cwd: Optional[str] = None
    ) -> ShellCommand:
        return ShellCommand(
            self, cmd, callback, std_err_callback, env, log_stderr, cwd
        )

    def run_cmd_async(
        self,
        cmd: str,
        callback: OutputCallback = None,
        std_err_callback: OutputCallback = None,
        timeout: float = 2.,
        attempts: int = 1,
        verbose: bool = True,
        sig_idx: int = 1,
        proc_input: Optional[str] = None,
        log_complete: bool = True,
        log_stderr: bool = False,
        env: Optional[Dict[str, str]] = None,
        cwd: Optional[str] = None,
        success_codes: Optional[List[int]] = None
    ) -> Awaitable[None]:
        """
        Runs a command and processes responses as they are received. Optional
        callbacks may be provided to handle stdout and stderr.
        """
        scmd = ShellCommand(
            self, cmd, callback, std_err_callback, env, log_stderr, cwd
        )
        attempts = max(1, attempts)
        async def _wrapper() -> None:
            for _ in range(attempts):
                if await scmd.run(
                    timeout, verbose, log_complete, sig_idx,
                    proc_input, success_codes
                ):
                    break
            else:
                ret_code = scmd.get_return_code()
                raise ShellCommandError(f"Error running command {cmd}", ret_code)
        return asyncio.create_task(_wrapper())

    def exec_cmd(
        self,
        cmd: str,
        timeout: float = 2.,
        attempts: int = 1,
        sig_idx: int = 1,
        proc_input: Optional[str] = None,
        log_complete: bool = True,
        log_stderr: bool = False,
        env: Optional[Dict[str, str]] = None,
        cwd: Optional[str] = None,
        success_codes: Optional[List[int]] = None
    ) -> Awaitable[str]:
        """
        Executes a command and returns UTF-8 decoded stdout upon completion.
        """
        scmd = ShellCommand(self, cmd, None, None, env,
                            log_stderr, cwd)
        coro = scmd.run_with_response(
            timeout, attempts, log_complete, sig_idx,
            proc_input, success_codes
        )
        return asyncio.create_task(coro)

    async def close(self) -> None:
        for cmd in self.running_commands:
            await cmd.cancel()

def load_component(config: ConfigHelper) -> ShellCommandFactory:
    return ShellCommandFactory(config)



# Async file locking using flock
#
# Copyright (C) 2024 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import os
import fcntl
import errno
import logging
import pathlib
import contextlib
import asyncio
from . import ServerError
from typing import Optional, Type, Union
from types import TracebackType

class LockTimeout(ServerError):
    pass

class AsyncExclusiveFileLock(contextlib.AbstractAsyncContextManager):
    def __init__(
        self, file_path: pathlib.Path, timeout: Union[int, float] = 0
    ) -> None:
        self.lock_path = file_path.parent.joinpath(f".{file_path.name}.lock")
        self.timeout = timeout
        self.fd: int = -1
        self.locked: bool = False
        self.required_wait: bool = False

    async def __aenter__(self) -> AsyncExclusiveFileLock:
        await self.acquire()
        return self

    async def __aexit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType]
    ) -> None:
        await self.release()

    def _get_lock(self) -> bool:
        flags = os.O_RDWR | os.O_CREAT | os.O_TRUNC
        fd = os.open(str(self.lock_path), flags, 0o644)
        with contextlib.suppress(PermissionError):
            os.chmod(fd, 0o644)
        try:
            fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except OSError as err:
            os.close(fd)
            if err.errno == errno.ENOSYS:
                raise
            return False
        stat = os.fstat(fd)
        if stat.st_nlink == 0:
            # File was deleted before opening and after acquiring
            # lock, create a new one
            os.close(fd)
            return False
        self.fd = fd
        return True

    async def acquire(self) -> None:
        self.required_wait = False
        if self.timeout < 0:
            return
        loop = asyncio.get_running_loop()
        endtime = loop.time() + self.timeout
        logged: bool = False
        while True:
            try:
                self.locked = await loop.run_in_executor(None, self._get_lock)
            except OSError as err:
                logging.info(
                    "Failed to aquire advisory lock, allowing unlocked entry."
                    f"Error: {err}"
                )
                self.locked = False
                return
            if self.locked:
                return
            self.required_wait = True
            await asyncio.sleep(.25)
            if not logged:
                logged = True
                logging.info(
                    f"File lock {self.lock_path} is currently acquired by another "
                    "process, waiting for release."
                )
            if self.timeout > 0 and endtime >= loop.time():
                raise LockTimeout(
                    f"Attempt to acquire lock '{self.lock_path}' timed out"
                )

    def _release_file(self) -> None:
        with contextlib.suppress(OSError, PermissionError):
            if self.lock_path.is_file():
                self.lock_path.unlink()
        with contextlib.suppress(OSError, PermissionError):
            fcntl.flock(self.fd, fcntl.LOCK_UN)
        with contextlib.suppress(OSError, PermissionError):
            os.close(self.fd)

    async def release(self) -> None:
        if not self.locked:
            return
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._release_file)
        self.locked = False



from __future__ import annotations
import pytest
import pytest_asyncio
import asyncio
import socket
import pathlib
from collections import namedtuple

from moonraker.server import CORE_COMPONENTS, Server, API_VERSION
from moonraker.server import main as servermain
from moonraker.eventloop import EventLoop
from moonraker.utils import ServerError
from moonraker.confighelper import ConfigError
from moonraker.components.klippy_apis import KlippyAPI
from mocks import MockComponent, MockWebsocket

from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Dict,
    Optional
)

if TYPE_CHECKING:
    from fixtures import HttpClient, WebsocketClient

MockArgs = namedtuple('MockArgs', ["logfile", "nologfile", "configfile"])

@pytest.mark.run_paths(moonraker_conf="invalid_config.conf")
def test_invalid_config(path_args: Dict[str, pathlib.Path]):
    evtloop = EventLoop()
    args = {
        'config_file': str(path_args['moonraker.conf']),
        'log_file': "",
        'software_version': "moonraker-pytest"
    }
    with pytest.raises(ConfigError):
        Server(args, None, evtloop)

def test_config_and_log_warnings(path_args: Dict[str, pathlib.Path]):
    evtloop = EventLoop()
    args = {
        'config_file': str(path_args['moonraker.conf']),
        'log_file': "",
        'software_version': "moonraker-pytest",
        'log_warning': "Log Warning Test",
        'config_warning': "Config Warning Test"
    }
    expected = ["Log Warning Test", "Config Warning Test"]
    server = Server(args, None, evtloop)
    assert server.warnings == expected

@pytest.mark.run_paths(moonraker_conf="unparsed_server.conf")
@pytest.mark.asyncio
async def test_unparsed_config_items(full_server: Server):
    expected_warnings = [
        "Unparsed config section [machine unparsed] detected.",
        "Unparsed config option 'unknown_option: True' detected "
        "in section [server]."]
    warn_cnt = 0
    for warn in full_server.warnings:
        for expected in expected_warnings:
            if warn.startswith(expected):
                warn_cnt += 1
    assert warn_cnt == 2

@pytest.mark.run_paths(moonraker_log="moonraker.log")
@pytest.mark.asyncio
async def test_file_logger(base_server: Server,
                           path_args: Dict[str, pathlib.Path]):
    log_path = path_args.get("moonraker.log", None)
    assert log_path is not None and log_path.exists()

def test_signal_handler(base_server: Server,
                        event_loop: asyncio.AbstractEventLoop):
    base_server._handle_term_signal()
    event_loop.run_forever()
    assert base_server.exit_reason == "terminate"

class TestInstantiation:
    def test_running(self, base_server: Server):
        assert base_server.is_running() is False

    def test_app_args(self,
                      path_args: Dict[str, pathlib.Path],
                      base_server: Server):
        args = {
            'config_file': str(path_args['moonraker.conf']),
            'log_file': str(path_args.get("moonlog", "")),
            'software_version': "moonraker-pytest"
        }
        assert base_server.get_app_args() == args

    def test_api_version(self, base_server: Server):
        ver = base_server.get_api_version()
        assert ver == API_VERSION

    def test_pending_tasks(self, base_server: Server):
        loop = base_server.get_event_loop().aioloop
        assert len(asyncio.all_tasks(loop)) == 0

    def test_klippy_info(self, base_server: Server):
        assert base_server.get_klippy_info() == {}

    def test_klippy_state(self, base_server: Server):
        assert base_server.get_klippy_state() == "disconnected"

    def test_host_info(self, base_server: Server):
        hinfo = {
            'hostname': socket.gethostname(),
            'address': "0.0.0.0",
            'port': 7010,
            'ssl_port': 7011
        }
        assert base_server.get_host_info() == hinfo

    def test_klippy_connection(self, base_server: Server):
        assert base_server.klippy_connection.is_connected() is False

    def test_components(self, base_server: Server):
        key_list = sorted(list(base_server.components.keys()))
        assert key_list == [
            "application",
            "internal_transport",
            "klippy_connection",
            "websockets",
        ]

    def test_endpoint_registered(self, base_server: Server):
        app = base_server.moonraker_app
        assert "/server/info" in app.api_cache

    @pytest.mark.asyncio
    async def test_notification(self, base_server: Server):
        base_server.register_notification("test:test_event")
        fut = base_server.event_loop.create_future()
        wsm = base_server.lookup_component("websockets")
        wsm.websockets[1] = MockWebsocket(fut)
        base_server.send_event("test:test_event", "test")
        ret = await fut
        expected = {
            'jsonrpc': "2.0",
            'method': "notify_test_event",
            'params': ["test"]
        }
        assert expected == ret

class TestLoadComponent:
    def test_load_component_fail(self, base_server: Server):
        with pytest.raises(ServerError):
            base_server.load_component(
                base_server.config, "invalid_component")

    def test_failed_component_set(self, base_server: Server):
        assert "invalid_component" in base_server.failed_components

    def test_load_component_fail_with_default(self, base_server: Server):
        comp = base_server.load_component(
            base_server.config, "invalid_component", None)
        assert comp is None

    def test_lookup_failed(self, base_server: Server):
        with pytest.raises(ServerError):
            base_server.lookup_component("invalid_component")

    def test_lookup_failed_with_default(self, base_server: Server):
        comp = base_server.lookup_component("invalid_component", None)
        assert comp is None

    def test_load_component(self, base_server: Server):
        comp = base_server.load_component(base_server.config, "klippy_apis")
        assert isinstance(comp, KlippyAPI)

    def test_lookup_component(self, base_server: Server):
        comp = base_server.lookup_component('klippy_apis')
        assert isinstance(comp, KlippyAPI)

    def test_component_attr(self, base_server: Server):
        key_list = sorted(list(base_server.components.keys()))
        assert key_list == [
            "application",
            "internal_transport",
            "klippy_apis",
            "klippy_connection",
            "websockets",
        ]

class TestCoreServer:
    @pytest_asyncio.fixture(scope="class")
    async def core_server(self, base_server: Server) -> AsyncIterator[Server]:
        base_server.load_components()
        yield base_server
        await base_server._stop_server("terminate")

    def test_running(self, core_server: Server):
        assert core_server.is_running() is False

    def test_http_servers(self, core_server: Server):
        app = core_server.lookup_component("application")
        assert (
            app.http_server is None and
            app.secure_server is None
        )

    def test_warnings(self, core_server: Server):
        assert len(core_server.warnings) == 0

    def test_failed_components(self, core_server: Server):
        assert len(core_server.failed_components) == 0

    def test_lookup_components(self, core_server: Server):
        comps = []
        for comp_name in CORE_COMPONENTS:
            comps.append(core_server.lookup_component(comp_name, None))
        assert None not in comps

    def test_pending_tasks(self, core_server: Server):
        loop = core_server.get_event_loop().aioloop
        assert len(asyncio.all_tasks(loop)) == 0

    def test_register_component_fail(self, core_server: Server):
        with pytest.raises(ServerError):
            core_server.register_component("machine", object())

    def test_register_remote_method(self, core_server: Server):
        core_server.register_remote_method("moonraker_test", lambda: None)
        kconn = core_server.klippy_connection
        assert "moonraker_test" in kconn.remote_methods

    def test_register_method_exists(self, core_server: Server):
        with pytest.raises(ServerError):
            core_server.register_remote_method(
                "shutdown_machine", lambda: None)

class TestServerInit:
    def test_running(self, full_server: Server):
        assert full_server.is_running() is False

    def test_http_servers(self, full_server: Server):
        app = full_server.lookup_component("application")
        assert (
            app.http_server is None and
            app.secure_server is None
        )

    def test_warnings(self, full_server: Server):
        assert len(full_server.warnings) == 0

    def test_failed_components(self, full_server: Server):
        assert len(full_server.failed_components) == 0

    def test_lookup_components(self, full_server: Server):
        comps = []
        for comp_name in CORE_COMPONENTS:
            comps.append(full_server.lookup_component(comp_name, None))
        assert None not in comps

    def test_config_backup(self,
                           full_server: Server,
                           path_args: Dict[str, pathlib.Path]):
        cfg = path_args["config_path"].joinpath(".moonraker.conf.bkp")
        assert cfg.is_file()

class TestServerStart:
    @pytest_asyncio.fixture(scope="class")
    async def server(self, full_server: Server) -> Server:
        await full_server.start_server(connect_to_klippy=False)
        return full_server

    def test_running(self, server: Server):
        assert server.is_running() is True

    def test_http_servers(self, server: Server):
        app = server.lookup_component("application")
        assert (
            app.http_server is not None and
            app.secure_server is None
        )

@pytest.mark.run_paths(moonraker_conf="base_server_ssl.conf")
class TestSecureServerStart:
    @pytest_asyncio.fixture(scope="class")
    async def server(self, full_server: Server) -> Server:
        await full_server.start_server(connect_to_klippy=False)
        return full_server

    def test_running(self, server: Server):
        assert server.is_running() is True

    def test_http_servers(self, server: Server):
        app = server.lookup_component("application")
        assert (
            app.http_server is not None and
            app.secure_server is not None
        )

@pytest.mark.asyncio
async def test_component_init_error(base_server: Server):
    base_server.register_component("testcomp", MockComponent(err_init=True))
    await base_server.server_init(False)
    assert "testcomp" in base_server.failed_components

@pytest.mark.asyncio
async def test_component_exit_error(base_server: Server,
                                    caplog: pytest.LogCaptureFixture):
    base_server.register_component("testcomp", MockComponent(err_exit=True))
    await base_server._stop_server("terminate")
    expected = "Error executing 'on_exit()' for component: testcomp"
    assert expected in caplog.messages

@pytest.mark.asyncio
async def test_component_close_error(base_server: Server,
                                     caplog: pytest.LogCaptureFixture):
    base_server.register_component("testcomp", MockComponent(err_close=True))
    await base_server._stop_server("terminate")
    expected = "Error executing 'close()' for component: testcomp"
    assert expected in caplog.messages

def test_register_event(base_server: Server):
    def test_func():
        pass
    base_server.register_event_handler("test:my_test", test_func)
    assert base_server.events["test:my_test"] == [test_func]

def test_register_async_event(base_server: Server):
    async def test_func():
        pass
    base_server.register_event_handler("test:my_test", test_func)
    assert base_server.events["test:my_test"] == [test_func]

@pytest.mark.asyncio
async def test_send_event(full_server: Server):
    evtloop = full_server.get_event_loop()
    fut = evtloop.create_future()

    def test_func(arg):
        fut.set_result(arg)
    full_server.register_event_handler("test:my_test", test_func)
    full_server.send_event("test:my_test", "test")
    result = await fut
    assert result == "test"

@pytest.mark.asyncio
async def test_send_async_event(full_server: Server):
    evtloop = full_server.get_event_loop()
    fut = evtloop.create_future()

    async def test_func(arg):
        fut.set_result(arg)
    full_server.register_event_handler("test:my_test", test_func)
    full_server.send_event("test:my_test", "test")
    result = await fut
    assert result == "test"

@pytest.mark.asyncio
async def test_register_remote_method_running(full_server: Server):
    await full_server.start_server(connect_to_klippy=False)
    with pytest.raises(ServerError):
        full_server.register_remote_method(
            "moonraker_test", lambda: None)

@pytest.mark.usefixtures("event_loop")
def test_main(path_args: Dict[str, pathlib.Path],
              monkeypatch: pytest.MonkeyPatch,
              caplog: pytest.LogCaptureFixture):
    tries = [1]

    def mock_init(self: Server):
        reason = "terminate"
        if tries:
            reason = "restart"
            tries.pop(0)
        self.event_loop.delay_callback(.01, self._stop_server, reason)
    cfg_path = path_args["moonraker.conf"]
    args = MockArgs("", True, str(cfg_path))
    monkeypatch.setattr(Server, "server_init", mock_init)
    code: Optional[int] = None
    try:
        servermain(args)
    except SystemExit as e:
        code = e.code
    assert (
        code == 0 and
        "Attempting Server Restart..." in caplog.messages and
        "Server Shutdown" == caplog.messages[-1]
    )

@pytest.mark.run_paths(moonraker_conf="invalid_config.conf")
def test_main_config_error(path_args: Dict[str, pathlib.Path],
                           caplog: pytest.LogCaptureFixture):
    cfg_path = path_args["moonraker.conf"]
    args = MockArgs("", True, str(cfg_path))
    try:
        servermain(args)
    except SystemExit as e:
        code = e.code
    assert code == 1 and "Server Config Error" in caplog.messages

@pytest.mark.run_paths(moonraker_conf="invalid_config.conf",
                       moonraker_bkp=".moonraker.conf.bkp")
@pytest.mark.usefixtures("event_loop")
def test_main_restore_config(path_args: Dict[str, pathlib.Path],
                             monkeypatch: pytest.MonkeyPatch,
                             caplog: pytest.LogCaptureFixture):
    def mock_init(self: Server):
        reason = "terminate"
        self.event_loop.delay_callback(.01, self._stop_server, reason)

    cfg_path = path_args["moonraker.conf"]
    args = MockArgs("", True, str(cfg_path))
    monkeypatch.setattr(Server, "server_init", mock_init)
    code: Optional[int] = None
    try:
        servermain(args)
    except SystemExit as e:
        code = e.code
    assert (
        code == 0 and
        "Loaded server from most recent working configuration:" in caplog.text
    )

class TestEndpoints:
    @pytest_asyncio.fixture(scope="class")
    async def server(self, full_server: Server):
        await full_server.start_server()
        yield full_server

    @pytest.mark.asyncio
    async def test_http_server_info(self,
                                    server: Server,
                                    http_client: HttpClient):
        ret = await http_client.get("/server/info")
        comps = list(server.components.keys())
        expected = {
            'klippy_connected': False,
            'klippy_state': "disconnected",
            'components': comps,
            'failed_components': [],
            'registered_directories': ["config", "logs"],
            'warnings': [],
            'websocket_count': 0,
            'moonraker_version': "moonraker-pytest",
            'missing_klippy_requirements': [],
            'api_version': list(API_VERSION),
            'api_version_string': ".".join(str(v) for v in API_VERSION)
        }
        assert ret["result"] == expected

    @pytest.mark.asyncio
    async def test_http_server_config(self,
                                      server: Server,
                                      http_client: HttpClient):
        cfg = server.config.get_parsed_config()
        ret = await http_client.get("/server/config")
        assert ret["result"]["config"] == cfg

    @pytest.mark.asyncio
    async def test_websocket_server_info(self,
                                         server: Server,
                                         websocket_client: WebsocketClient):
        ret = await websocket_client.request("server.info")
        comps = list(server.components.keys())
        expected = {
            'klippy_connected': False,
            'klippy_state': "disconnected",
            'components': comps,
            'failed_components': [],
            'registered_directories': ["config", "logs"],
            'warnings': [],
            'websocket_count': 1,
            'moonraker_version': "moonraker-pytest",
            'missing_klippy_requirements': [],
            'api_version': list(API_VERSION),
            'api_version_string': ".".join(str(v) for v in API_VERSION)
        }
        assert ret == expected

    @pytest.mark.asyncio
    async def test_websocket_server_config(self,
                                           server: Server,
                                           websocket_client: WebsocketClient):
        cfg = server.config.get_parsed_config()
        ret = await websocket_client.request("server.config")
        assert ret["config"] == cfg

def test_server_restart(base_server: Server,
                        http_client: HttpClient,
                        event_loop: asyncio.AbstractEventLoop):
    result = {}

    async def do_restart():
        base_server.load_components()
        await base_server.start_server()
        ret = await http_client.post("/server/restart")
        result.update(ret)
    event_loop.create_task(do_restart())
    event_loop.run_forever()
    assert result["result"] == "ok" and base_server.exit_reason == "restart"

@pytest.mark.no_ws_connect
def test_websocket_restart(base_server: Server,
                           websocket_client: WebsocketClient,
                           event_loop: asyncio.AbstractEventLoop):
    result = {}

    async def do_restart():
        base_server.load_components()
        await base_server.start_server()
        await websocket_client.connect()
        ret = await websocket_client.request("server.restart")
        result["result"] = ret
    event_loop.create_task(do_restart())
    event_loop.run_forever()
    assert result["result"] == "ok" and base_server.exit_reason == "restart"


# TODO:
# test invalid cert, key (probably should do that in test_app.py)



# Generic sensor support
#
# Copyright (C) 2022 Morton Jonuschat <mjonuschat+moonraker@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

# Component to read additional generic sensor data and make it
# available to clients
from __future__ import annotations

import logging
from collections import defaultdict, deque
from functools import partial
from ..common import RequestType, HistoryFieldData

# Annotation imports
from typing import (
    Any,
    DefaultDict,
    Deque,
    Dict,
    List,
    Optional,
    Type,
    TYPE_CHECKING,
    Union,
    Callable
)

if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .mqtt import MQTTClient
    from .history import History

SENSOR_UPDATE_TIME = 1.0
SENSOR_EVENT_NAME = "sensors:sensor_update"

def _set_result(
    name: str, value: Union[int, float], store: Dict[str, Union[int, float]]
) -> None:
    if not isinstance(value, (int, float)):
        store[name] = float(value)
    else:
        store[name] = value


class BaseSensor:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.error_state: Optional[str] = None
        self.id = config.get_name().split(maxsplit=1)[-1]
        self.type = config.get("type")
        self.name = config.get("name", self.id)
        self.last_measurements: Dict[str, Union[int, float]] = {}
        self.last_value: Dict[str, Union[int, float]] = {}
        store_size = config.getint("sensor_store_size", 1200)
        self.values: DefaultDict[str, Deque[Union[int, float]]] = defaultdict(
            lambda: deque(maxlen=store_size)
        )
        self.param_info: List[Dict[str, str]] = []
        history: History = self.server.lookup_component("history")
        self.field_info: Dict[str, List[HistoryFieldData]] = {}
        all_opts = list(config.get_options().keys())
        cfg_name = config.get_name()
        param_prefix = "parameter_"
        hist_field_prefix = "history_field_"
        for opt in all_opts:
            if opt.startswith(param_prefix):
                name = opt[len(param_prefix):]
                data = config.getdict(opt)
                data["name"] = opt[len(param_prefix):]
                self.param_info.append(data)
                continue
            if not opt.startswith(hist_field_prefix):
                continue
            name = opt[len(hist_field_prefix):]
            field_cfg: Dict[str, str] = config.getdict(opt)
            ident: Optional[str] = field_cfg.pop("parameter", None)
            if ident is None:
                raise config.error(
                    f"[{cfg_name}]: option '{opt}', key 'parameter' must be"
                    f"specified"
                )
            do_init: str = field_cfg.pop("init_tracker", "false").lower()
            reset_cb = self._gen_reset_callback(ident) if do_init == "true" else None
            excl_paused: str = field_cfg.pop("exclude_paused", "false").lower()
            report_total: str = field_cfg.pop("report_total", "false").lower()
            report_max: str = field_cfg.pop("report_maximum", "false").lower()
            precision: Optional[str] = field_cfg.pop("precision", None)
            try:
                fdata = HistoryFieldData(
                    name,
                    cfg_name,
                    field_cfg.pop("desc", f"{ident} tracker"),
                    field_cfg.pop("strategy", "basic"),
                    units=field_cfg.pop("units", None),
                    reset_callback=reset_cb,
                    exclude_paused=excl_paused == "true",
                    report_total=report_total == "true",
                    report_maximum=report_max == "true",
                    precision=int(precision) if precision is not None else None,
                )
            except Exception as e:
                raise config.error(
                    f"[{cfg_name}]: option '{opt}', error encountered during "
                    f"sensor field configuration: {e}"
                ) from e
            for key in field_cfg.keys():
                self.server.add_warning(
                    f"[{cfg_name}]: Option '{opt}' contains invalid key '{key}'"
                )
            self.field_info.setdefault(ident, []).append(fdata)
            history.register_auxiliary_field(fdata)

    def _gen_reset_callback(self, param_name: str) -> Callable[[], float]:
        def on_reset() -> float:
            return self.last_measurements.get(param_name, 0)
        return on_reset

    def _update_sensor_value(self, eventtime: float) -> None:
        """
        Append the last updated value to the store.
        """
        for key, value in self.last_measurements.items():
            self.values[key].append(value)

        # Copy the last measurements data
        self.last_value = {**self.last_measurements}

    async def initialize(self) -> bool:
        """
        Sensor initialization executed on Moonraker startup.
        """
        logging.info("Registered sensor '%s'", self.name)
        return True

    def get_sensor_info(self, extended: bool = False) -> Dict[str, Any]:
        ret: Dict[str, Any] = {
            "id": self.id,
            "friendly_name": self.name,
            "type": self.type,
            "values": self.last_measurements,
        }
        if extended:
            ret["parameter_info"] = self.param_info
            history_fields: List[Dict[str, Any]] = []
            for parameter, field_list in self.field_info.items():
                for field_data in field_list:
                    field_config = field_data.get_configuration()
                    field_config["parameter"] = parameter
                    history_fields.append(field_config)
            ret["history_fields"] = history_fields
        return ret

    def get_sensor_measurements(self) -> Dict[str, List[Union[int, float]]]:
        return {key: list(values) for key, values in self.values.items()}

    def get_name(self) -> str:
        return self.name

    def close(self) -> None:
        pass


class MQTTSensor(BaseSensor):
    def __init__(self, config: ConfigHelper) -> None:
        super().__init__(config=config)
        self.mqtt: MQTTClient = self.server.load_component(config, "mqtt")
        self.state_topic: str = config.get("state_topic")
        self.state_response = config.gettemplate("state_response_template")
        self.qos: Optional[int] = config.getint("qos", None, minval=0, maxval=2)
        self.server.register_event_handler(
            "mqtt:disconnected", self._on_mqtt_disconnected
        )

    def _on_state_update(self, payload: bytes) -> None:
        measurements: Dict[str, Union[int, float]] = {}
        context = {
            "payload": payload.decode(),
            "set_result": partial(_set_result, store=measurements),
            "log_debug": logging.debug
        }

        try:
            self.state_response.render(context)
        except Exception as e:
            logging.error("Error updating sensor results: %s", e)
            self.error_state = str(e)
        else:
            self.error_state = None
            self.last_measurements = measurements
            for name, value in measurements.items():
                fdata_list = self.field_info.get(name)
                if fdata_list is None:
                    continue
                for fdata in fdata_list:
                    fdata.tracker.update(value)

    async def _on_mqtt_disconnected(self):
        self.error_state = "MQTT Disconnected"
        self.last_measurements = {}

    async def initialize(self) -> bool:
        await super().initialize()
        try:
            self.mqtt.subscribe_topic(
                self.state_topic,
                self._on_state_update,
                self.qos,
            )
            self.error_state = None
            return True
        except Exception as e:
            self.error_state = str(e)
            return False


class Sensors:
    __sensor_types: Dict[str, Type[BaseSensor]] = {"MQTT": MQTTSensor}

    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.sensors: Dict[str, BaseSensor] = {}

        # Register timer to update sensor values in store
        self.sensors_update_timer = self.server.get_event_loop().register_timer(
            self._update_sensor_values
        )

        # Register endpoints
        self.server.register_endpoint(
            "/server/sensors/list",
            RequestType.GET,
            self._handle_sensor_list_request,
        )
        self.server.register_endpoint(
            "/server/sensors/info",
            RequestType.GET,
            self._handle_sensor_info_request,
        )
        self.server.register_endpoint(
            "/server/sensors/measurements",
            RequestType.GET,
            self._handle_sensor_measurements_request,
        )

        # Register notifications
        self.server.register_notification(SENSOR_EVENT_NAME)
        prefix_sections = config.get_prefix_sections("sensor ")
        for section in prefix_sections:
            cfg = config[section]
            try:
                try:
                    _, name = cfg.get_name().split(maxsplit=1)
                except ValueError:
                    raise cfg.error(f"Invalid section name: {cfg.get_name()}")
                logging.info(f"Configuring sensor: {name}")
                sensor_type: str = cfg.get("type")
                sensor_class: Optional[Type[BaseSensor]] = self.__sensor_types.get(
                    sensor_type.upper(), None
                )
                if sensor_class is None:
                    raise config.error(f"Unsupported sensor type: {sensor_type}")

                self.sensors[name] = sensor_class(cfg)
            except Exception as e:
                # Ensures that configuration errors are shown to the user
                self.server.add_warning(
                    f"Failed to configure sensor [{cfg.get_name()}]\n{e}", exc_info=e
                )
                continue

    def _update_sensor_values(self, eventtime: float) -> float:
        """
        Iterate through the sensors and store the last updated value.
        """
        changed_data: Dict[str, Dict[str, Union[int, float]]] = {}
        for sensor_name, sensor in self.sensors.items():
            base_value = sensor.last_value
            sensor._update_sensor_value(eventtime=eventtime)

            # Notify if a change in sensor values was detected
            if base_value != sensor.last_value:
                changed_data[sensor_name] = sensor.last_value
        if changed_data:
            self.server.send_event(SENSOR_EVENT_NAME, changed_data)
        return eventtime + SENSOR_UPDATE_TIME

    async def component_init(self) -> None:
        try:
            logging.debug("Initializing sensor component")
            for sensor in self.sensors.values():
                if not await sensor.initialize():
                    self.server.add_warning(
                        f"Sensor '{sensor.get_name()}' failed to initialize"
                    )

            self.sensors_update_timer.start()
        except Exception as e:
            logging.exception(e)

    async def _handle_sensor_list_request(
        self, web_request: WebRequest
    ) -> Dict[str, Dict[str, Any]]:
        extended = web_request.get_boolean("extended", False)
        return {
            "sensors": {
                key: sensor.get_sensor_info(extended)
                for key, sensor in self.sensors.items()
            }
        }

    async def _handle_sensor_info_request(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        sensor_name: str = web_request.get_str("sensor")
        extended = web_request.get_boolean("extended", False)
        if sensor_name not in self.sensors:
            raise self.server.error(f"No valid sensor named {sensor_name}")
        sensor = self.sensors[sensor_name]
        return sensor.get_sensor_info(extended)

    async def _handle_sensor_measurements_request(
        self, web_request: WebRequest
    ) -> Dict[str, Dict[str, Any]]:
        sensor_name: str = web_request.get_str("sensor", "")
        if sensor_name:
            sensor = self.sensors.get(sensor_name, None)
            if sensor is None:
                raise self.server.error(f"No valid sensor named {sensor_name}")
            return {sensor_name: sensor.get_sensor_measurements()}
        else:
            return {
                key: sensor.get_sensor_measurements()
                for key, sensor in self.sensors.items()
            }

    def close(self) -> None:
        self.sensors_update_timer.stop()
        for sensor in self.sensors.values():
            sensor.close()


def load_component(config: ConfigHelper) -> Sensors:
    return Sensors(config)



# Provides System Package Updates
#
# Copyright (C) 2023  Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import asyncio
import logging
import time
import re
from ...thirdparty.packagekit import enums as PkEnum
from .base_deploy import BaseDeploy

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Optional,
    Union,
    Dict,
    List,
)

if TYPE_CHECKING:
    from ...confighelper import ConfigHelper
    from ..shell_command import ShellCommandFactory as SCMDComp
    from ..dbus_manager import DbusManager
    from ..machine import Machine
    from .update_manager import CommandHelper
    from dbus_next import Variant
    from dbus_next.aio import ProxyInterface
    JsonType = Union[List[Any], Dict[str, Any]]


class PackageDeploy(BaseDeploy):
    def __init__(self,
                 config: ConfigHelper,
                 cmd_helper: CommandHelper
                 ) -> None:
        super().__init__(config, cmd_helper, "system", "", "")
        cmd_helper.set_package_updater(self)
        self.use_packagekit = config.getboolean("enable_packagekit", True)
        self.available_packages: List[str] = []

    async def initialize(self) -> Dict[str, Any]:
        storage = await super().initialize()
        self.available_packages = storage.get('packages', [])
        provider: BasePackageProvider
        try_fallback = True
        if self.use_packagekit:
            try:
                provider = PackageKitProvider(self.cmd_helper)
                await provider.initialize()
            except Exception:
                pass
            else:
                self.log_info("PackageDeploy: PackageKit Provider Configured")
                self.prefix = "PackageKit: "
                try_fallback = False
        if try_fallback:
            # Check to see of the apt command is available
            fallback = await self._get_fallback_provider()
            if fallback is None:
                provider = BasePackageProvider(self.cmd_helper)
                machine: Machine = self.server.lookup_component("machine")
                dist_info = machine.get_system_info()['distribution']
                dist_id: str = dist_info['id'].lower()
                self.server.add_warning(
                    "Unable to initialize System Update Provider for "
                    f"distribution: {dist_id}")
            else:
                self.log_info("PackageDeploy: Using APT CLI Provider")
                self.prefix = "Package Manager APT: "
                provider = fallback
        self.provider = provider
        return storage

    async def _get_fallback_provider(self) -> Optional[BasePackageProvider]:
        # Currently only the API Fallback provider is available
        shell_cmd: SCMDComp
        shell_cmd = self.server.lookup_component("shell_command")
        cmd = shell_cmd.build_shell_command("sh -c 'command -v apt'")
        try:
            ret = await cmd.run_with_response()
        except shell_cmd.error:
            return None
        # APT Command found should be available
        self.log_debug(f"APT package manager detected: {ret}")
        provider = AptCliProvider(self.cmd_helper)
        try:
            await provider.initialize()
        except Exception:
            return None
        return provider

    async def refresh(self) -> None:
        try:
            # Do not force a refresh until the server has started
            if self.server.is_running():
                await self._update_package_cache(force=True)
            self.available_packages = await self.provider.get_packages()
            pkg_msg = "\n".join(self.available_packages)
            self.log_info(
                f"Detected {len(self.available_packages)} package updates:"
                f"\n{pkg_msg}"
            )
        except Exception:
            self.log_exc("Error Refreshing System Packages")
        # Update Persistent Storage
        self._save_state()

    def get_persistent_data(self) -> Dict[str, Any]:
        storage = super().get_persistent_data()
        storage['packages'] = self.available_packages
        return storage

    async def update(self) -> bool:
        if not self.available_packages:
            return False
        self.cmd_helper.notify_update_response("Updating packages...")
        try:
            await self._update_package_cache(force=True, notify=True)
            await self.provider.upgrade_system()
        except Exception:
            raise self.server.error("Error updating system packages")
        self.available_packages = []
        self._save_state()
        self.cmd_helper.notify_update_response(
            "Package update finished...", is_complete=True)
        return True

    async def _update_package_cache(self,
                                    force: bool = False,
                                    notify: bool = False
                                    ) -> None:
        curtime = time.time()
        if force or curtime > self.last_refresh_time + 3600.:
            # Don't update if a request was done within the last hour
            await self.provider.refresh_packages(notify)

    async def install_packages(self,
                               package_list: List[str],
                               **kwargs
                               ) -> None:
        await self.provider.install_packages(package_list, **kwargs)

    def get_update_status(self) -> Dict[str, Any]:
        return {
            'package_count': len(self.available_packages),
            'package_list': self.available_packages
        }

class BasePackageProvider:
    def __init__(self, cmd_helper: CommandHelper) -> None:
        self.server = cmd_helper.get_server()
        self.cmd_helper = cmd_helper

    async def initialize(self) -> None:
        pass

    async def refresh_packages(self, notify: bool = False) -> None:
        raise self.server.error("Cannot refresh packages, no provider set")

    async def get_packages(self) -> List[str]:
        raise self.server.error("Cannot retrieve packages, no provider set")

    async def install_packages(self,
                               package_list: List[str],
                               **kwargs
                               ) -> None:
        raise self.server.error("Cannot install packages, no provider set")

    async def upgrade_system(self) -> None:
        raise self.server.error("Cannot upgrade packages, no provider set")

class AptCliProvider(BasePackageProvider):
    APT_CMD = "sudo DEBIAN_FRONTEND=noninteractive apt-get"

    async def refresh_packages(self, notify: bool = False) -> None:
        await self.cmd_helper.run_cmd(
            f"{self.APT_CMD} update", timeout=600., notify=notify)

    async def get_packages(self) -> List[str]:
        shell_cmd = self.cmd_helper.get_shell_command()
        res = await shell_cmd.exec_cmd("apt list --upgradable", timeout=60.)
        pkg_list = [p.strip() for p in res.split("\n") if p.strip()]
        if pkg_list:
            pkg_list = pkg_list[2:]
            return [p.split("/", maxsplit=1)[0] for p in pkg_list]
        return []

    async def resolve_packages(self, package_list: List[str]) -> List[str]:
        self.cmd_helper.notify_update_response("Resolving packages...")
        search_regex = "|".join([f"^{pkg}$" for pkg in package_list])
        cmd = f"apt-cache search --names-only \"{search_regex}\""
        shell_cmd = self.cmd_helper.get_shell_command()
        ret = await shell_cmd.exec_cmd(cmd, timeout=600.)
        resolved = [
            pkg.strip().split()[0] for pkg in ret.split("\n") if pkg.strip()
        ]
        return [avail for avail in package_list if avail in resolved]

    async def install_packages(self,
                               package_list: List[str],
                               **kwargs
                               ) -> None:
        timeout: float = kwargs.get('timeout', 300.)
        retries: int = kwargs.get('retries', 3)
        notify: bool = kwargs.get('notify', False)
        await self.refresh_packages(notify=notify)
        resolved = await self.resolve_packages(package_list)
        if not resolved:
            self.cmd_helper.notify_update_response("No packages detected")
            return
        logging.debug(f"Resolved packages: {resolved}")
        pkgs = " ".join(resolved)
        await self.cmd_helper.run_cmd(
            f"{self.APT_CMD} install --yes {pkgs}", timeout=timeout,
            attempts=retries, notify=notify)

    async def upgrade_system(self) -> None:
        await self.cmd_helper.run_cmd(
            f"{self.APT_CMD} upgrade --yes", timeout=3600.,
            notify=True)

class PackageKitProvider(BasePackageProvider):
    def __init__(self, cmd_helper: CommandHelper) -> None:
        super().__init__(cmd_helper)
        dbus_mgr: DbusManager = self.server.lookup_component("dbus_manager")
        self.dbus_mgr = dbus_mgr
        self.pkgkit: Optional[ProxyInterface] = None

    async def initialize(self) -> None:
        if not self.dbus_mgr.is_connected():
            raise self.server.error("DBus Connection Not available")
        # Check for PolicyKit permissions
        await self.dbus_mgr.check_permission(
            "org.freedesktop.packagekit.system-sources-refresh",
            "The Update Manager will fail to fetch package updates")
        await self.dbus_mgr.check_permission(
            "org.freedesktop.packagekit.package-install",
            "The Update Manager will fail to install packages")
        await self.dbus_mgr.check_permission(
            "org.freedesktop.packagekit.system-update",
            "The Update Manager will fail to update packages"
        )
        # Fetch the PackageKit DBus Inteface
        self.pkgkit = await self.dbus_mgr.get_interface(
            "org.freedesktop.PackageKit",
            "/org/freedesktop/PackageKit",
            "org.freedesktop.PackageKit")

    async def refresh_packages(self, notify: bool = False) -> None:
        await self.run_transaction("refresh_cache", False, notify=notify)

    async def get_packages(self) -> List[str]:
        flags = PkEnum.Filter.NONE
        pkgs = await self.run_transaction("get_updates", flags.value)
        pkg_ids = [info['package_id'] for info in pkgs if 'package_id' in info]
        return [pkg_id.split(";")[0] for pkg_id in pkg_ids]

    async def install_packages(self,
                               package_list: List[str],
                               **kwargs
                               ) -> None:
        notify: bool = kwargs.get('notify', False)
        await self.refresh_packages(notify=notify)
        flags = (
            PkEnum.Filter.NEWEST | PkEnum.Filter.NOT_INSTALLED |
            PkEnum.Filter.BASENAME | PkEnum.Filter.ARCH
        )
        pkgs = await self.run_transaction("resolve", flags.value, package_list)
        pkg_ids = [info['package_id'] for info in pkgs if 'package_id' in info]
        if pkg_ids:
            logging.debug(f"Installing Packages: {pkg_ids}")
            tflag = PkEnum.TransactionFlag.ONLY_TRUSTED
            await self.run_transaction("install_packages", tflag.value,
                                       pkg_ids, notify=notify)

    async def upgrade_system(self) -> None:
        # Get Updates, Install Packages
        flags = PkEnum.Filter.NONE
        pkgs = await self.run_transaction("get_updates", flags.value)
        pkg_ids = [info['package_id'] for info in pkgs if 'package_id' in info]
        if pkg_ids:
            logging.debug(f"Upgrading Packages: {pkg_ids}")
            tflag = PkEnum.TransactionFlag.ONLY_TRUSTED
            await self.run_transaction("update_packages", tflag.value,
                                       pkg_ids, notify=True)

    def create_transaction(self) -> PackageKitTransaction:
        if self.pkgkit is None:
            raise self.server.error("PackageKit Interface Not Available")
        return PackageKitTransaction(self.dbus_mgr, self.pkgkit,
                                     self.cmd_helper)

    async def run_transaction(self,
                              method: str,
                              *args,
                              notify: bool = False
                              ) -> Any:
        transaction = self.create_transaction()
        return await transaction.run(method, *args, notify=notify)

class PackageKitTransaction:
    GET_PKG_ROLES = (
        PkEnum.Role.RESOLVE | PkEnum.Role.GET_PACKAGES |
        PkEnum.Role.GET_UPDATES
    )
    QUERY_ROLES = GET_PKG_ROLES | PkEnum.Role.GET_REPO_LIST
    PROGRESS_STATUS = (
        PkEnum.Status.RUNNING | PkEnum.Status.INSTALL |
        PkEnum.Status.UPDATE
    )

    def __init__(self,
                 dbus_mgr: DbusManager,
                 pkgkit: ProxyInterface,
                 cmd_helper: CommandHelper
                 ) -> None:
        self.server = cmd_helper.get_server()
        self.eventloop = self.server.get_event_loop()
        self.cmd_helper = cmd_helper
        self.dbus_mgr = dbus_mgr
        self.pkgkit = pkgkit
        # Transaction Properties
        self.notify = False
        self._status = PkEnum.Status.UNKNOWN
        self._role = PkEnum.Role.UNKNOWN
        self._tflags = PkEnum.TransactionFlag.NONE
        self._percentage = 101
        self._dl_remaining = 0
        self.speed = 0
        self.elapsed_time = 0
        self.remaining_time = 0
        self.caller_active = False
        self.allow_cancel = True
        self.uid = 0
        # Transaction data tracking
        self.tfut: Optional[asyncio.Future] = None
        self.last_progress_notify_time: float = 0.
        self.result: List[Dict[str, Any]] = []
        self.err_msg: str = ""

    def run(self,
            method: str,
            *args,
            notify: bool = False
            ) -> Awaitable:
        if self.tfut is not None:
            raise self.server.error(
                "PackageKit transaction can only be used once")
        self.notify = notify
        self.tfut = self.eventloop.create_future()
        coro = self._start_transaction(method, *args)
        self.eventloop.create_task(coro)
        return self.tfut

    async def _start_transaction(self,
                                 method: str,
                                 *args
                                 ) -> None:
        assert self.tfut is not None
        try:
            # Create Transaction
            tid = await self.pkgkit.call_create_transaction()  # type: ignore
            transaction, props = await self.dbus_mgr.get_interfaces(
                "org.freedesktop.PackageKit", tid,
                ["org.freedesktop.PackageKit.Transaction",
                 "org.freedesktop.DBus.Properties"])
            # Set interface callbacks
            transaction.on_package(self._on_package_signal)    # type: ignore
            transaction.on_repo_detail(                        # type: ignore
                self._on_repo_detail_signal)
            transaction.on_item_progress(                      # type: ignore
                self._on_item_progress_signal)
            transaction.on_error_code(self._on_error_signal)   # type: ignore
            transaction.on_finished(self._on_finished_signal)  # type: ignore
            props.on_properties_changed(                       # type: ignore
                self._on_properties_changed)
            # Run method
            logging.debug(f"PackageKit: Running transaction call_{method}")
            func = getattr(transaction, f"call_{method}")
            await func(*args)
        except Exception as e:
            self.tfut.set_exception(e)

    def _on_package_signal(self,
                           info_code: int,
                           package_id: str,
                           summary: str
                           ) -> None:
        info = PkEnum.Info.from_index(info_code)
        if self._role in self.GET_PKG_ROLES:
            pkg_data = {
                'package_id': package_id,
                'info': info.desc,
                'summary': summary
            }
            self.result.append(pkg_data)
        else:
            self._notify_package(info, package_id)

    def _on_repo_detail_signal(self,
                               repo_id: str,
                               description: str,
                               enabled: bool
                               ) -> None:
        if self._role == PkEnum.Role.GET_REPO_LIST:
            repo_data = {
                "repo_id": repo_id,
                "description": description,
                "enabled": enabled
            }
            self.result.append(repo_data)
        else:
            self._notify_repo(repo_id, description)

    def _on_item_progress_signal(self,
                                 item_id: str,
                                 status_code: int,
                                 percent_complete: int
                                 ) -> None:
        status = PkEnum.Status.from_index(status_code)  # noqa: F841
        # NOTE: This signal doesn't seem to fire predictably,
        # nor does it seem to provide a consistent "percent complete"
        # parameter.
        # logging.debug(
        #    f"Role {self._role.name}: Item Progress Signal Received\n"
        #    f"Item ID: {item_id}\n"
        #    f"Percent Complete: {percent_complete}\n"
        #    f"Status: {status.desc}")

    def _on_error_signal(self,
                         error_code: int,
                         details: str
                         ) -> None:
        err = PkEnum.Error.from_index(error_code)
        self.err_msg = f"{err.name}: {details}"

    def _on_finished_signal(self, exit_code: int, run_time: int) -> None:
        if self.tfut is None:
            return
        ext = PkEnum.Exit.from_index(exit_code)
        secs = run_time / 1000.
        if ext == PkEnum.Exit.SUCCESS:
            self.tfut.set_result(self.result)
        else:
            err = self.err_msg or ext.desc
            server = self.cmd_helper.get_server()
            self.tfut.set_exception(server.error(err))
        msg = f"Transaction {self._role.desc}: Exit {ext.desc}, " \
              f"Run time: {secs:.2f} seconds"
        if self.notify:
            self.cmd_helper.notify_update_response(msg)
        logging.debug(msg)

    def _on_properties_changed(self,
                               iface_name: str,
                               changed_props: Dict[str, Variant],
                               invalid_props: Dict[str, Variant]
                               ) -> None:
        for name, var in changed_props.items():
            formatted = re.sub(r"(\w)([A-Z])", r"\g<1>_\g<2>", name).lower()
            setattr(self, formatted, var.value)

    def _notify_package(self, info: PkEnum.Info, package_id: str) -> None:
        if self.notify:
            if info == PkEnum.Info.FINISHED:
                return
            pkg_parts = package_id.split(";")
            msg = f"{info.desc}: {pkg_parts[0]} ({pkg_parts[1]})"
            self.cmd_helper.notify_update_response(msg)

    def _notify_repo(self, repo_id: str, description: str) -> None:
        if self.notify:
            if not repo_id.strip():
                repo_id = description
            # TODO: May want to eliminate dups
            msg = f"GET: {repo_id}"
            self.cmd_helper.notify_update_response(msg)

    def _notify_progress(self) -> None:
        if self.notify and self._percentage <= 100:
            msg = f"{self._status.desc}...{self._percentage}%"
            if self._status == PkEnum.Status.DOWNLOAD and self._dl_remaining:
                if self._dl_remaining < 1024:
                    msg += f", Remaining: {self._dl_remaining} B"
                elif self._dl_remaining < 1048576:
                    msg += f", Remaining: {self._dl_remaining // 1024} KiB"
                else:
                    msg += f", Remaining: {self._dl_remaining // 1048576} MiB"
                if self.speed:
                    speed = self.speed // 8
                    if speed < 1024:
                        msg += f", Speed: {speed} B/s"
                    elif speed < 1048576:
                        msg += f", Speed: {speed // 1024} KiB/s"
                    else:
                        msg += f", Speed: {speed // 1048576} MiB/s"
            self.cmd_helper.notify_update_response(msg)

    @property
    def role(self) -> PkEnum.Role:
        return self._role

    @role.setter
    def role(self, role_code: int) -> None:
        self._role = PkEnum.Role.from_index(role_code)
        if self._role in self.QUERY_ROLES:
            # Never Notify Queries
            self.notify = False
        if self.notify:
            msg = f"Transaction {self._role.desc} started..."
            self.cmd_helper.notify_update_response(msg)
        logging.debug(f"PackageKit: Current Role: {self._role.desc}")

    @property
    def status(self) -> PkEnum.Status:
        return self._status

    @status.setter
    def status(self, status_code: int) -> None:
        self._status = PkEnum.Status.from_index(status_code)
        self._percentage = 101
        self.speed = 0
        logging.debug(f"PackageKit: Current Status: {self._status.desc}")

    @property
    def transaction_flags(self) -> PkEnum.TransactionFlag:
        return self._tflags

    @transaction_flags.setter
    def transaction_flags(self, bits: int) -> None:
        self._tflags = PkEnum.TransactionFlag(bits)

    @property
    def percentage(self) -> int:
        return self._percentage

    @percentage.setter
    def percentage(self, percent: int) -> None:
        self._percentage = percent
        if self._status in self.PROGRESS_STATUS:
            self._notify_progress()

    @property
    def download_size_remaining(self) -> int:
        return self._dl_remaining

    @download_size_remaining.setter
    def download_size_remaining(self, bytes_remaining: int) -> None:
        self._dl_remaining = bytes_remaining
        self._notify_progress()



# Async CAN Socket utility
#
# Copyright (C) 2023 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import socket
import asyncio
import errno
import struct
import logging
from . import ServerError
from typing import List, Dict, Optional, Union

CAN_FMT = "<IB3x8s"
CAN_READER_LIMIT = 1024 * 1024
KLIPPER_ADMIN_ID = 0x3f0
KLIPPER_SET_NODE_CMD = 0x01
KATAPULT_SET_NODE_CMD = 0x11
CMD_QUERY_UNASSIGNED = 0x00
CANBUS_RESP_NEED_NODEID = 0x20

class CanNode:
    def __init__(self, node_id: int, cansocket: CanSocket) -> None:
        self.node_id = node_id
        self._reader = asyncio.StreamReader(CAN_READER_LIMIT)
        self._cansocket = cansocket

    async def read(
        self, n: int = -1, timeout: Optional[float] = 2
    ) -> bytes:
        return await asyncio.wait_for(self._reader.read(n), timeout)

    async def readexactly(
        self, n: int, timeout: Optional[float] = 2
    ) -> bytes:
        return await asyncio.wait_for(self._reader.readexactly(n), timeout)

    async def readuntil(
        self, sep: bytes = b"\x03", timeout: Optional[float] = 2
    ) -> bytes:
        return await asyncio.wait_for(self._reader.readuntil(sep), timeout)

    def write(self, payload: Union[bytes, bytearray]) -> None:
        if isinstance(payload, bytearray):
            payload = bytes(payload)
        self._cansocket.send(self.node_id, payload)

    async def write_with_response(
        self,
        payload: Union[bytearray, bytes],
        resp_length: int,
        timeout: Optional[float] = 2.
    ) -> bytes:
        self.write(payload)
        return await self.readexactly(resp_length, timeout)

    def feed_data(self, data: bytes) -> None:
        self._reader.feed_data(data)

    def close(self) -> None:
        self._reader.feed_eof()

class CanSocket:
    def __init__(self, interface: str):
        self._loop = asyncio.get_running_loop()
        self.nodes: Dict[int, CanNode] = {}
        self.cansock = socket.socket(socket.PF_CAN, socket.SOCK_RAW, socket.CAN_RAW)
        self.input_buffer = b""
        self.output_packets: List[bytes] = []
        self.input_busy = False
        self.output_busy = False
        self.closed = True
        try:
            self.cansock.bind((interface,))
        except Exception:
            raise ServerError(f"Unable to bind socket to interface '{interface}'", 500)
        self.closed = False
        self.cansock.setblocking(False)
        self._loop.add_reader(self.cansock.fileno(), self._handle_can_response)

    def register_node(self, node_id: int) -> CanNode:
        if node_id in self.nodes:
            return self.nodes[node_id]
        node = CanNode(node_id, self)
        self.nodes[node_id + 1] = node
        return node

    def remove_node(self, node_id: int) -> None:
        node = self.nodes.pop(node_id + 1, None)
        if node is not None:
            node.close()

    def _handle_can_response(self) -> None:
        try:
            data = self.cansock.recv(4096)
        except socket.error as e:
            # If bad file descriptor allow connection to be
            # closed by the data check
            if e.errno == errno.EBADF:
                logging.exception("Can Socket Read Error, closing")
                data = b''
            else:
                return
        if not data:
            # socket closed
            self.close()
            return
        self.input_buffer += data
        if self.input_busy:
            return
        self.input_busy = True
        while len(self.input_buffer) >= 16:
            packet = self.input_buffer[:16]
            self._process_packet(packet)
            self.input_buffer = self.input_buffer[16:]
        self.input_busy = False

    def _process_packet(self, packet: bytes) -> None:
        can_id, length, data = struct.unpack(CAN_FMT, packet)
        can_id &= socket.CAN_EFF_MASK
        payload = data[:length]
        node = self.nodes.get(can_id)
        if node is not None:
            node.feed_data(payload)

    def send(self, can_id: int, payload: bytes = b"") -> None:
        if can_id > 0x7FF:
            can_id |= socket.CAN_EFF_FLAG
        if not payload:
            packet = struct.pack(CAN_FMT, can_id, 0, b"")
            self.output_packets.append(packet)
        else:
            while payload:
                length = min(len(payload), 8)
                pkt_data = payload[:length]
                payload = payload[length:]
                packet = struct.pack(
                    CAN_FMT, can_id, length, pkt_data)
                self.output_packets.append(packet)
        if self.output_busy:
            return
        self.output_busy = True
        asyncio.create_task(self._do_can_send())

    async def _do_can_send(self):
        while self.output_packets:
            packet = self.output_packets.pop(0)
            try:
                await self._loop.sock_sendall(self.cansock, packet)
            except socket.error:
                logging.info("Socket Write Error, closing")
                self.close()
                break
        self.output_busy = False

    def close(self):
        if self.closed:
            return
        self.closed = True
        for node in self.nodes.values():
            node.close()
        self._loop.remove_reader(self.cansock.fileno())
        self.cansock.close()

async def query_klipper_uuids(can_socket: CanSocket) -> List[Dict[str, str]]:
    loop = asyncio.get_running_loop()
    admin_node = can_socket.register_node(KLIPPER_ADMIN_ID)
    payload = bytes([CMD_QUERY_UNASSIGNED])
    admin_node.write(payload)
    curtime = loop.time()
    endtime = curtime + 2.
    uuids: List[Dict[str, str]] = []
    while curtime < endtime:
        timeout = max(.1, endtime - curtime)
        try:
            resp = await admin_node.read(8, timeout)
        except asyncio.TimeoutError:
            continue
        finally:
            curtime = loop.time()
        if len(resp) < 7 or resp[0] != CANBUS_RESP_NEED_NODEID:
            continue
        app_names = {
            KLIPPER_SET_NODE_CMD: "Klipper",
            KATAPULT_SET_NODE_CMD: "Katapult"
        }
        app = "Unknown"
        if len(resp) > 7:
            app = app_names.get(resp[7], "Unknown")
        data = resp[1:7]
        uuids.append(
            {
                "uuid": data.hex(),
                "application": app
            }
        )
    return uuids



from __future__ import annotations
import pathlib
import pytest
import hashlib
import confighelper
import shutil
import time
from moonraker.confighelper import ConfigError
from moonraker.server import Server
from moonraker.utils import ServerError
from moonraker.components import gpio
from mocks import MockGpiod
from typing import TYPE_CHECKING, Dict
if TYPE_CHECKING:
    from confighelper import ConfigHelper

@pytest.fixture(scope="class")
def config(base_server: Server) -> ConfigHelper:
    base_server.load_component(base_server.config, "secrets")
    return base_server.config

@pytest.fixture(scope="class")
def test_config(config: ConfigHelper,
                path_args: Dict[str, pathlib.Path]
                ) -> ConfigHelper:
    assets = path_args['asset_path']
    sup_cfg_path = assets.joinpath("moonraker/supplemental.conf")
    if not sup_cfg_path.exists():
        pytest.fail("Supplemental config not found")
    cfg = config.read_supplemental_config(str(sup_cfg_path))
    return cfg["test_options"]

@pytest.fixture(scope="function")
def gpio_config(test_config: ConfigHelper,
                monkeypatch: pytest.MonkeyPatch
                ) -> ConfigHelper:
    def load_gpio_mock(name: str) -> MockGpiod:
        return MockGpiod()
    monkeypatch.setattr(gpio, "load_system_module", load_gpio_mock)
    yield test_config
    server = test_config.get_server()
    gpio_comp = server.lookup_component("gpio", None)
    if gpio_comp is not None:
        gpio_comp.close()
        gpio_comp.reserved_gpios = {}

class TestConfigGeneric:
    def test_get_server(self, config: ConfigHelper):
        server = config.get_server()
        assert isinstance(server, Server)

    def test_get_item(self, config: ConfigHelper):
        sec = config["file_manager"]
        assert sec.section == "file_manager"

    def test_no_section_fail(self, config: ConfigHelper):
        with pytest.raises(ConfigError):
            config["not_available"].get("no_section")

    def test_contains(self, config: ConfigHelper):
        assert "file_manager" in config

    def test_not_contains(self, config: ConfigHelper):
        assert "not_available" not in config

    def test_has_option(self, config: ConfigHelper):
        assert config.has_option("host")

    def test_get_name(self, config: ConfigHelper):
        assert config.get_name() == "server"

    def test_get_options(self,
                         config: ConfigHelper,
                         path_args: Dict[str, pathlib.Path]):
        expected = {
            "host": "0.0.0.0",
            "port": "7010",
            "ssl_port": "7011",
            "klippy_uds_address": str(path_args["klippy_uds_path"])
        }
        assert expected == config.get_options()

    def test_get_hash(self, config: ConfigHelper):
        opts = config.get_options()
        expected_hash = hashlib.sha256()
        for opt, val in opts.items():
            expected_hash.update(opt.encode())
            expected_hash.update(val.encode())
        cfg_hash = config.get_hash().hexdigest()
        assert cfg_hash == expected_hash.hexdigest()

def test_missing_supplemental_config(config: ConfigHelper):
    no_file = pathlib.Path("nofile")
    with pytest.raises(ConfigError):
        config.read_supplemental_config(no_file)

def test_error_supplemental_config(config: ConfigHelper,
                                   path_args: Dict[str, pathlib.Path]):
    assets = path_args["asset_path"]
    invalid_cfg = assets.joinpath("moonraker/invalid_config.conf")
    if not invalid_cfg.exists():
        pytest.fail("Invalid Config File does not exist")
    with pytest.raises(ConfigError):
        config.read_supplemental_config(invalid_cfg)

def test_prefix_sections(test_config: ConfigHelper):
    prefix = test_config.get_prefix_sections("prefix_sec")
    expected = ["prefix_sec one", "prefix_sec two", "prefix_sec three"]
    assert prefix == expected

class TestGetString:
    def test_get_str_exists(self, test_config: ConfigHelper):
        val = test_config.get("test_string")
        assert val == "Hello World"

    def test_get_str_fail(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.get("invalid_option")

    def test_get_str_default(self, test_config: ConfigHelper):
        assert test_config.get("invalid_option", None) is None

    def test_get_str_deprecate(self, test_config: ConfigHelper):
        server = test_config.get_server()
        test_config.get("test_string", deprecate=True)
        expected = (
            f"[test_options]: Option 'test_string' is "
            "deprecated, see the configuration documention "
            "at https://moonraker.readthedocs.io/en/latest/configuration"
        )
        assert expected in server.warnings

class TestGetInt:
    def test_get_int_exists(self, test_config: ConfigHelper):
        val = test_config.getint("test_int")
        assert val == 1

    def test_get_int_fail(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getint("invalid_option")

    def test_get_int_default(self, test_config: ConfigHelper):
        assert test_config.getint("invalid_option", None) is None

    def test_get_int_fail_above(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getint("test_int", above=1)

    def test_get_int_fail_below(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getint("test_int", below=1)

    def test_get_int_fail_minval(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getint("test_int", minval=2)

    def test_get_int_fail_maxval(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getint("test_int", maxval=0)

    def test_get_int_pass_all(self, test_config: ConfigHelper):
        val = test_config.getint("test_int", above=0, below=2,
                                 minval=1, maxval=1)
        assert val == 1

    def test_get_int_deprecate(self, test_config: ConfigHelper):
        server = test_config.get_server()
        test_config.getint("test_int", deprecate=True)
        expected = (
            f"[test_options]: Option 'test_int' is "
            "deprecated, see the configuration documention "
            "at https://moonraker.readthedocs.io/en/latest/configuration"
        )
        assert expected in server.warnings

class TestGetFloat:
    def test_get_float_exists(self, test_config: ConfigHelper):
        val = test_config.getfloat("test_float")
        assert 3.5 == pytest.approx(val)

    def test_get_float_fail(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getfloat("invalid_option")

    def test_get_float_default(self, test_config: ConfigHelper):
        assert test_config.getfloat("invalid_option", None) is None

    def test_get_float_fail_above(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getfloat("test_float", above=3.55)

    def test_get_float_fail_below(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getfloat("test_float", below=3.45)

    def test_get_float_fail_minval(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getfloat("test_float", minval=3.6)

    def test_get_float_fail_maxval(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getfloat("test_float", maxval=3.45)

    def test_get_float_pass_all(self, test_config: ConfigHelper):
        val = test_config.getfloat("test_float", above=3.45, below=3.55,
                                   minval=3, maxval=4)
        assert 3.5 == pytest.approx(val)

    def test_get_float_deprecate(self, test_config: ConfigHelper):
        server = test_config.get_server()
        test_config.getfloat("test_float", deprecate=True)
        expected = (
            f"[test_options]: Option 'test_float' is "
            "deprecated, see the configuration documention "
            "at https://moonraker.readthedocs.io/en/latest/configuration"
        )
        assert expected in server.warnings

class TestGetBoolean:
    def test_get_boolean_exists(self, test_config: ConfigHelper):
        val = test_config.getboolean("test_bool")
        assert val is True

    def test_get_float_fail(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getboolean("invalid_option")

    def test_get_float_default(self, test_config: ConfigHelper):
        assert test_config.getboolean("invalid_option", None) is None

    def test_get_int_deprecate(self, test_config: ConfigHelper):
        server = test_config.get_server()
        test_config.getboolean("test_bool", deprecate=True)
        expected = (
            f"[test_options]: Option 'test_bool' is "
            "deprecated, see the configuration documention "
            "at https://moonraker.readthedocs.io/en/latest/configuration"
        )
        assert expected in server.warnings

class TestGetList:
    def test_get_list_exists(self, test_config: ConfigHelper):
        val = test_config.getlist("test_list")
        assert val == ["one", "two", "three"]

    def test_get_list_fail(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getlist("invalid_option")

    def test_get_list_default(self, test_config: ConfigHelper):
        assert test_config.getlist("invalid_option", None) is None

    def test_get_int_list(self, test_config: ConfigHelper):
        val = test_config.getintlist("test_int_list", separator=",")
        assert val == [1, 2, 3]

    def test_get_float_list(self, test_config: ConfigHelper):
        val = test_config.getfloatlist("test_float_list", separator=",")
        assert val == pytest.approx([1.5, 2.8, 3.2])

    def test_get_multi_list(self, test_config: ConfigHelper):
        val = test_config.getlists("test_multi_list", list_type=int,
                                   separators=("\n", ","))
        assert val == [[1, 2, 3], [4, 5, 6]]

    def test_get_list_deprecate(self, test_config: ConfigHelper):
        server = test_config.get_server()
        test_config.getlist("test_list", deprecate=True)
        expected = (
            f"[test_options]: Option 'test_list' is "
            "deprecated, see the configuration documention "
            "at https://moonraker.readthedocs.io/en/latest/configuration"
        )
        assert expected in server.warnings

class TestGetDict:
    def test_get_dict_exists(self, test_config: ConfigHelper):
        val = test_config.getdict("test_dict", dict_type=int)
        assert val == {"one": 1, "two": 2, "three": 3}

    def test_get_dict_fail(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getdict("invalid_option")

    def test_get_dict_default(self, test_config: ConfigHelper):
        assert test_config.getdict("invalid_option", None) is None

    def test_get_dict_empty_fields(self, test_config: ConfigHelper):
        val = test_config.getdict("test_dict_empty_field",
                                  allow_empty_fields=True)
        assert val == {"one": "test", "two": None, "three": None}

    def test_get_dict_empty_fields_fail(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.getdict("test_dict_empty_field")

    def test_get_dict_deprecate(self, test_config: ConfigHelper):
        server = test_config.get_server()
        test_config.getdict("test_dict", deprecate=True)
        expected = (
            f"[test_options]: Option 'test_dict' is "
            "deprecated, see the configuration documention "
            "at https://moonraker.readthedocs.io/en/latest/configuration"
        )
        assert expected in server.warnings

class TestGetTemplate:
    def test_get_template_exists(self, test_config: ConfigHelper):
        val = test_config.gettemplate("test_template").render()
        assert val == "mqttuser"

    @pytest.mark.asyncio
    async def test_get_template_async(self, test_config: ConfigHelper):
        templ = test_config.gettemplate("test_template", is_async=True)
        val = await templ.render_async()
        assert val == "mqttuser"

    def test_get_template_plain(self, test_config: ConfigHelper):
        val = test_config.gettemplate("test_string").render()
        assert val == "Hello World"

    def test_get_template_fail(self, test_config: ConfigHelper):
        with pytest.raises(ConfigError):
            test_config.gettemplate("invalid_option")

    def test_get_template_render_fail(self, test_config: ConfigHelper):
        with pytest.raises(ServerError):
            test_config.gettemplate("test_template", is_async=True).render()

    def test_get_template_default(self, test_config: ConfigHelper):
        assert test_config.gettemplate("invalid_option", None) is None

    def test_load_template(self, test_config: ConfigHelper):
        val = test_config.load_template("test_template").render()
        assert val == "mqttuser"

    def test_load_template_default(self, test_config: ConfigHelper):
        templ = test_config.load_template(
            "invalid_option", "{secrets.mqtt_credentials.password}")
        val = templ.render()
        assert val == "mqttpass"

    def test_get_template_deprecate(self, test_config: ConfigHelper):
        server = test_config.get_server()
        test_config.gettemplate("test_template", deprecate=True)
        expected = (
            f"[test_options]: Option 'test_template' is "
            "deprecated, see the configuration documention "
            "at https://moonraker.readthedocs.io/en/latest/configuration"
        )
        assert expected in server.warnings

class TestGetGpioOut:
    def test_get_gpio_exists(self, gpio_config: ConfigHelper):
        val: gpio.GpioOutputPin = gpio_config.getgpioout("test_gpio")
        assert (
            val.orig == "gpiochip0/gpio26" and
            val.name == "gpiochip0:gpio26" and
            val.inverted is False and
            val.value == 0
        )

    def test_get_gpio_no_chip(self, gpio_config: ConfigHelper):
        val: gpio.GpioOutputPin = gpio_config.getgpioout("test_gpio_no_chip")
        assert (
            val.orig == "gpio26" and
            val.name == "gpiochip0:gpio26" and
            val.inverted is False and
            val.value == 0
        )

    def test_get_gpio_invert(self, gpio_config: ConfigHelper):
        val: gpio.GpioOutputPin = gpio_config.getgpioout("test_gpio_invert")
        assert (
            val.orig == "!gpiochip0/gpio26" and
            val.name == "gpiochip0:gpio26" and
            val.inverted is True and
            val.value == 0
        )

    def test_get_gpio_no_chip_invert(self, gpio_config: ConfigHelper):
        val: gpio.GpioOutputPin = gpio_config.getgpioout(
            "test_gpio_no_chip_invert")
        assert (
            val.orig == "!gpio26" and
            val.name == "gpiochip0:gpio26" and
            val.inverted is True and
            val.value == 0
        )

    def test_get_gpio_initial_value(self, gpio_config: ConfigHelper):
        val: gpio.GpioOutputPin = gpio_config.getgpioout(
            "test_gpio", initial_value=1)
        assert (
            val.orig == "gpiochip0/gpio26" and
            val.name == "gpiochip0:gpio26" and
            val.inverted is False and
            val.value == 1
        )

    def test_get_gpio_fail(self, gpio_config: ConfigHelper):
        with pytest.raises(ConfigError):
            gpio_config.getgpioout("invalid_option")

    def test_get_gpio_default(self, gpio_config: ConfigHelper):
        assert gpio_config.getgpioout("invalid_option", None) is None

    @pytest.mark.parametrize("opt", ["pullup", "pullup_no_chip",
                                     "pulldown", "pulldown_no_chip"])
    def test_get_gpio_invalid(self, gpio_config: ConfigHelper, opt: str):
        option = f"test_gpio_{opt}"
        if not gpio_config.has_option(option):
            pytest.fail(f"No option {option}")
        with pytest.raises(ConfigError):
            gpio_config.getgpioout(option)

    def test_get_gpio_deprecated(self, gpio_config: ConfigHelper):
        server = gpio_config.get_server()
        gpio_config.getgpioout("test_gpio", deprecate=True)
        expected = (
            f"[test_options]: Option 'test_gpio' is "
            "deprecated, see the configuration documention "
            "at https://moonraker.readthedocs.io/en/latest/configuration"
        )
        assert expected in server.warnings

class TestGetConfiguration:
    def test_get_config_no_exist(self, base_server: Server):
        fake_path = pathlib.Path("no_exist")
        if fake_path.exists():
            pytest.fail("Path exists")
        args = dict(base_server.app_args)
        args["config_file"] = str(fake_path)
        with pytest.raises(ConfigError):
            confighelper.get_configuration(base_server, args)

    def test_get_config_no_access(self,
                                  base_server: Server,
                                  path_args: Dict[str, pathlib.Path]
                                  ):
        cfg_path = path_args["config_path"]
        test_cfg = cfg_path.joinpath("test.conf")
        shutil.copy(path_args["moonraker.conf"], test_cfg)
        test_cfg.chmod(mode=222)
        args = dict(base_server.app_args)
        args["config_file"] = str(test_cfg)
        with pytest.raises(ConfigError):
            confighelper.get_configuration(base_server, args)

    def test_get_config_no_server(self,
                                  base_server: Server,
                                  path_args: Dict[str, pathlib.Path]
                                  ):
        assets = path_args['asset_path']
        sup_cfg_path = assets.joinpath("moonraker/supplemental.conf")
        if not sup_cfg_path.exists():
            pytest.fail("Supplemental config not found")
        args = dict(base_server.app_args)
        args["config_file"] = str(sup_cfg_path)
        with pytest.raises(ConfigError):
            confighelper.get_configuration(base_server, args)

class TestBackupConfig:
    def test_find_backup_fail(self):
        fake_path = pathlib.Path("no_exist")
        if fake_path.exists():
            fake_path.unlink()
        result = confighelper.find_config_backup(fake_path)
        assert result is None

    def test_backup_config_success(
        self, path_args: Dict[str, pathlib.Path], config: ConfigHelper
    ):
        cfg_path = path_args["moonraker.conf"]
        bkp_dest = cfg_path.parent.joinpath(f".{cfg_path.name}.bkp")
        if bkp_dest.exists():
            pytest.fail("Backup Already Exists")
        config.create_backup()
        assert bkp_dest.is_file()

    def test_backup_skip(
        self, path_args: Dict[str, pathlib.Path], config: ConfigHelper
    ):
        cfg_path = path_args["moonraker.conf"]
        bkp_dest = cfg_path.parent.joinpath(f".{cfg_path.name}.bkp")
        if not bkp_dest.exists():
            pytest.fail("Backup Not Present")
        stat = bkp_dest.stat()
        config.create_backup()
        assert stat == bkp_dest.stat()

    def test_find_backup(self, path_args: Dict[str, pathlib.Path]):
        cfg_path = path_args["moonraker.conf"]
        bkp_dest = cfg_path.parent.joinpath(f".{cfg_path.name}.bkp")
        bkp = confighelper.find_config_backup(str(cfg_path))
        assert bkp == str(bkp_dest)
